package explicit.graphviz;

import java.util.Collections;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

/**
 * A Decoration stores the label and additional attributes
 * for a GraphViz / DOT node or edge.
 * <br>
 * A default Decoration can be specified that serves to
 * fill in unspecified attributes (i.e., it will be output
 * as a default node / edge setting).
 */
public class Decoration
{
	/** The label */
	private String label;
	/** The additional attributes */
	private TreeMap<String, String> attributes;

	/** (Optional) A default decoration */ 
	private Decoration defaults;

	/** Constructor */
	public Decoration()
	{
		this.defaults = null;
		this.label = "";
	}

	/** Constructor, set default decoration */
	public Decoration(Decoration defaults)
	{
		this.defaults = defaults;
	}

	/** Set the label for this decoration */
	public void setLabel(String label)
	{
		this.label = label;
	}

	/** Set the label for this decoration */
	public String getLabel()
	{
		return label;
	}

	/**
	 * Read/write access to the (non-default) attributes
	 * of this decoration. 
	 */
	public Map<String, String> attributes()
	{
		if (attributes == null) {
			attributes = new TreeMap<String, String>();
		}
		return attributes;
	}

	/**
	 * Read-only access to the (non-default) attributes
	 * of this decoration.
	 * <br>
	 * If you only need read-only access, please use this
	 * method, as this avoids the allocation of a TreeSet
	 * for storing the attributes in the very common
	 * case that no additional attributes are present.
	 */
	public Map<String, String> attributesRO()
	{
		if (attributes == null) {
			return Collections.emptyMap();
		} else {
			return Collections.unmodifiableMap(attributes);
		}
	}

	/**
	 * Get the effective value for a given attribute key.
	 * The effective value is either the value specified
	 * in this decoration or the default value, if present.
	 *
	 * @param key the attribute key (may be "label" to access the label)
	 */
	public String getEffectiveValue(String key)
	{
		if (key.equals("label")) {
			return label;
		}

		String value = attributesRO().get(key);
		if (value == null && defaults != null) {
			return defaults.getEffectiveValue(key);
		}
		return value;
	}

	@Override
	public String toString()
	{
		StringBuffer buf = new StringBuffer();

		append(buf, "label", "\"" + label + "\"");

		for (Entry<String, String> e : attributesRO().entrySet()) {
			if (defaults != null) {
				String defaultValue = defaults.getEffectiveValue(e.getKey());
				if (defaultValue != null && defaultValue.equals(e.getValue())) {
					// skip, as it's the default value and we don't want to pollute the output
					continue;
				}
			}

			String value = e.getValue();
			value = "\"" + value + "\"";
			append(buf, e.getKey(), value);
		}

		if (buf.length() == 0) {
			return "";
		} else {
			return "[" + buf.toString() + "]";
		}
	}

	/**
	 * Append the given key and value to the buffer, prepending with at ','
	 * if necessary.
	 */
	protected void append(StringBuffer buffer, String key, String value)
	{
		if (buffer.length() != 0) {
			buffer.append(",");
		}
		buffer.append(key);
		buffer.append("=");
		buffer.append(value);
	}

}