package explicit.graphviz;

/**
 * A Decorator is an object that is called multiple
 * times during the output of a GraphViz (DOT) file
 * and can change the decoration (label and attributes)
 * of DOT nodes (for states) and DOT edges (for the various
 * kinds of transitions in a model).
 * <br>
 * Decorators can be applied one after another, changing
 * aspects of the decoration that will be ultimately
 * output in the DOT file.
 * <br>
 * Each method is given information about the current
 * context and the current decoration.
 * <br>
 * Decorators can assume that the label in the input
 * decoration is already set to a suitable string. 
 */
public interface Decorator
{
	/**
	 * This method is called to allow the decoration of
	 * a given state node to be changed.
	 * <br>
	 * This method may change the input decoration {@code d}.
	 * <br>
	 * Default implementation: Output decoration equals input decoration.
	 *
	 * @param state the state index
	 * @param d the current decoration
	 * @returns the new decoration (may be d, potentially modified)
	 */
	default Decoration decorateState(int state, Decoration d)
	{
		return d;
	}

	/**
	 * This method is called to allow the decoration of
	 * a non-deterministic choice edge to be changed.
	 * <br>
	 * This method may change the input decoration {@code d}.
	 * <br>
	 * Default implementation: Output decoration equals input decoration.
	 *
	 * @param from the state index of the originating state
	 * @param choice the choice index
	 * @param d the current decoration
	 * @returns the new decoration (may be d, potentially modified)
	 */
	default Decoration decorateTransition(int from, int choice, Decoration d)
	{
		return d;
	}

	/**
	 * This method is called to allow the decoration of
	 * the probabilistic edge in an MDP to be changed.
	 * <br>
	 * This method may change the input decoration {@code d}.
	 * <br>
	 * Default implementation: Output decoration equals input decoration.
	 *
	 * @param from the state index of the originating state
	 * @param to the state index of the destination state
	 * @param choice the choice index
	 * @param probability the probability (could be a Double, BigRational, a param.Function, ...)
	 * @param d the current decoration
	 * @returns the new decoration (may be d, potentially modified)
	 */
	default Decoration decorateProbability(int from, int to, int choice, Object probability, Decoration d)
	{
		return d;
	}

	/**
	 * This method is called to allow the decoration of
	 * the probabilistic edge in a DTMC/CTMC to be changed.
	 * <br>
	 * This method may change the input decoration {@code d}.
	 * <br>
	 * Default implementation: Output decoration equals input decoration.
	 *
	 * @param from the state index of the originating state
	 * @param to the state index of the destination state
	 * @param probability the probability/rate (could be a Double, BigRational, a param.Function, ...)
	 * @param d the current decoration
	 * @returns the new decoration (may be d, potentially modified)
	 */
	default Decoration decorateProbability(int from, int to, Object probability, Decoration d)
	{
		return d;
	}

	/**
	 * This method is called to allow the decoration of
	 * an automaton edge to be changed.
	 * <br>
	 * This method may change the input decoration {@code d}.
	 * <br>
	 * Default implementation: Output decoration equals input decoration.
	 *
	 * @param from the state index of the originating state
	 * @param to the state index of the destination state
	 * @param edgeLabel the edge label (could be String, BitSet, i.e., element of an APSet, ...)
	 * @param d the current decoration
	 * @returns the new decoration (may be d, potentially modified)
	 */
	default Decoration decorateAutomatonEdge(int from, int to, Object edgelabel, Decoration d)
	{
		return d;
	}

}
