package prism.conditional;

import java.io.File;
import java.io.FileNotFoundException;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.NavigableSet;
import java.util.Stack;
import java.util.TreeMap;
import java.util.TreeSet;

import common.SafeCast;
import common.StopWatch;
import common.functions.primitive.PairPredicateInt;
import common.iterable.IterableBitSet;
import explicit.MDPSparse;
import explicit.MinMax;
import explicit.SCCConsumerStore;
import explicit.modelviews.MDPDroppedChoices;
import explicit.rewards.MDPRewardsSimple;
import jdd.JDD;
import jdd.JDDNode;
import jdd.JDDVars;
import jdd.SanityJDD;
import mtbdd.PrismMTBDD;
import odd.ODDNode;
import odd.ODDUtils;
import parser.ast.Expression;
import parser.ast.ExpressionConditional;
import parser.ast.ExpressionLabel;
import parser.ast.ExpressionLiteral;
import parser.ast.ExpressionProb;
import parser.ast.ExpressionReward;
import parser.ast.ExpressionTemporal;
import parser.ast.RelOp;
import parser.ast.TemporalOperatorBound;
import parser.type.TypeInt;
import prism.ECComputer;
import prism.IntegerBound;
import prism.MDPEquiv;
import prism.ModelType;
import prism.MDPEquiv.MDPQuotient;
import prism.conditional.transform.MDPAddTauToFail;
import prism.conditional.transform.MDPAddTauToFail.MDPWithFail;
import prism.conditional.transform.MDPResetTransformation;
import prism.NondetModel;
import prism.NondetModelChecker;
import prism.NondetModelTransformation;
import prism.OpRelOpBound;
import prism.Prism;
import prism.PrismDevNullLog;
import prism.PrismException;
import prism.PrismFileLog;
import prism.PrismNotSupportedException;
import prism.PrismSettings;
import prism.PrismUtils;
import prism.ProbModel;
import prism.ProbModelChecker;
import prism.RewardCounterProduct;
import prism.SCCComputer;
import prism.StateValues;
import prism.StateValuesDV;
import prism.StateValuesMTBDD;

public class ConditionalExpectationsMDPModelChecker extends ConditionalModelChecker<NondetModel> {
	
	protected NondetModelChecker mc;
	private String failLabel;
	private String goalLabel;
	private boolean verbose = false;
	private boolean extraVerbose = false;

	private static final int TAU_CHOICE = -1;
	private static final int TRAP_CHOICE = -2;

	private static enum SaturationPointMethod { NAIVE, REWARD_QUANTILE, DIRECT, SKIP};
	private SaturationPointMethod saturationPointMethod;

	private static class ComputedValues {
		Constants consts;
		boolean isAcyclic = false;

		JDDNode actMax;
		/** Is ActMax = Act? */
		private boolean allActMax = false;

		JDDNode pmax;
		StateValuesMTBDD thetaSaturated;

		StateValuesMTBDD result;
		MDPQuotient mecQuotient;
		MDPWithFail cleanedUp;

		public ComputedValues()
		{
		}

		public ComputedValues(StateValuesMTBDD result)
		{
			this.result = result;
		}

		public void clear()
		{
			if (actMax != null) JDD.Deref(actMax);
			if (pmax != null) JDD.Deref(pmax);
			if (thetaSaturated != null) thetaSaturated.clear();

			if (result != null) result.clear();
			if (mecQuotient != null) mecQuotient.clear();
			if (cleanedUp != null) cleanedUp.clear();
		}
	}
	
	private static class Constants {
		/** Additional details used in the calculation of Lemma 6.17 */
		String details_6_17;

		Double p_6_35;

		Double lambda;
		Integer R0;
		Double C;
		Double delta_6_38;

		Integer Rturning;
		Double CEmaxUpperBound;
		Integer pSaturation;

		@Override
		public String toString()
		{
			return toString(false);
		}

		public String toString(boolean verbose)
		{
			String result = "R0 = " + (R0 == null ? "?" : R0);
			result += ", λ = " + (lambda == null ? "?" : lambda);
			result += ", C = " + (C == null ? "?" : C);
			result += ", Rturning = " + (Rturning == null ? "?" : Rturning);
			result += ", CE^ub = " + (CEmaxUpperBound == null ? "?" : CEmaxUpperBound);
			result += ", saturation point = " + (pSaturation == null ? "?" : pSaturation);
			if (verbose) {
				result += "\n   (" + (details_6_17 != null ? details_6_17+", "  : "");
				result += "p(6.35) = " + (p_6_35 == null ? "?" : p_6_35);
				result += ", delta(6.38) = " + (delta_6_38 == null ? "?" : delta_6_38);
				result += ")";
			}
			return result;
		}
	};

	public ConditionalExpectationsMDPModelChecker(NondetModelChecker mc, Prism prism) {
		super(prism);
		this.mc = mc;
		
		switch (prism.getSettings().getString(PrismSettings.CONDITIONAL_EXP_MDP_SATURATION_METHOD)) {
		case "naive":
		default:
			saturationPointMethod = SaturationPointMethod.NAIVE;
			break;
		case "quantile":
			saturationPointMethod = SaturationPointMethod.REWARD_QUANTILE;
			break;
		case "direct":
			saturationPointMethod = SaturationPointMethod.DIRECT;
			break;
		case "skip":
			saturationPointMethod = SaturationPointMethod.SKIP;
			break;
		}

		this.verbose = prism.getSettings().getBoolean(PrismSettings.CONDITIONAL_EXP_MDP_VERBOSE);
	}
	
	public StateValues checkExpression(final NondetModel model, final ExpressionConditional expression, JDDNode statesOfInterest) throws PrismException {
		final ExpressionReward objective = (ExpressionReward) expression.getObjective();
		OpRelOpBound oprel = objective.getRelopBoundInfo(mc.getConstantValues());

		if (!JDD.isSingleton(statesOfInterest, model.getAllDDRowVars())) {
			JDD.Deref(statesOfInterest);
			throw new PrismNotSupportedException("Currently, only a single state of interest is supported for MDP conditionals, got " + JDD.GetNumMintermsString(statesOfInterest, model.getNumDDRowVars()));
		}

		if (!model.getStart().equals(statesOfInterest)) {
			JDD.Deref(statesOfInterest);
			throw new PrismNotSupportedException("Currently, only a single, initial state of interest is supported for MDP conditional expectations");
			// TODO: handle general case, restrict to reachable part
		}

		if (oprel.getMinMax(ModelType.MDP).isMin()) {
			throw new PrismNotSupportedException("Currently, for conditional expectations in MDPs, only Rmax is supported");
		}

		// Get rewards
		Object rs = objective.getRewardStructIndex();
		JDDNode stateRewards = mc.getStateRewardsByIndexObject(rs, model, mc.getConstantValues());
		JDDNode transRewards = mc.getTransitionRewardsByIndexObject(rs, model, mc.getConstantValues());

		NondetModel simplified = new NondetModel(model.getTrans().copy(),
		                                         model.getStart().copy(),
		                                         new JDDNode[]{stateRewards.copy()},
		                                         new JDDNode[]{transRewards.copy()},
		                                         new String[]{"reward"},
		                                         model.getAllDDRowVars().copy(),
		                                         model.getAllDDColVars().copy(),
		                                         model.getAllDDSynchVars().copy(),
		                                         model.getAllDDSchedVars().copy(),
		                                         model.getAllDDChoiceVars().copy(),
		                                         model.getAllDDNondetVars().copy(),
		                                         model.getModelVariables().copy(),
		                                         model.getNumModules(),
		                                         model.getModuleNames(),
		                                         JDDVars.copyArray(model.getModuleDDRowVars()),
		                                         JDDVars.copyArray(model.getModuleDDColVars()),
		                                         model.getNumVars(),
		                                         model.getVarList(),
		                                         JDDVars.copyArray(model.getVarDDRowVars()),
		                                         JDDVars.copyArray(model.getVarDDColVars()),
		                                         model.getConstantValues());

		simplified.setSynchs(model.getSynchs());

		// TODO: currently, this is fine, as we have s_init = model.getStart();
		simplified.setReach(model.getReach().copy());
		simplified.filterReachableStates();

		StateValuesMTBDD result = null;
		try {
			StateValuesMTBDD resultSimplified = 
					computeValue(simplified,
					             (ExpressionReward)expression.getObjective(),
					             expression.getCondition(),
					             oprel.getMinMax(model.getModelType()).isMin()).convertToStateValuesMTBDD();

			result = new StateValuesMTBDD(resultSimplified.getJDDNode().copy(), model);
			resultSimplified.clear();
		} finally {
			simplified.clear();
			JDD.Deref(statesOfInterest);
		}

		return result;
	}
	
	private StateValues computeValue(final NondetModel model, final ExpressionReward objective, final Expression condition, boolean min) throws PrismException {
		StopWatch phase1 = new StopWatch(mc.getLog());
		phase1.start("first phase");

		// if we are here, model.getStart() is a singleton
		JDDNode sInit = model.getStart().copy();

		if (!(objective.getExpression() instanceof ExpressionTemporal) ||
		    !(((ExpressionTemporal)objective.getExpression()).getOperator() == ExpressionTemporal.R_F ||
		      ((ExpressionTemporal)objective.getExpression()).getOperator() == ExpressionTemporal.P_F)) {
			throw new PrismNotSupportedException("Only reachability objective for conditional expected reward");
		}
		if (((ExpressionTemporal)objective.getExpression()).hasBounds()) {
			throw new PrismNotSupportedException("Only simple reachability objective for conditional expected reward, no bounds");
		}
		Expression expressionF = ((ExpressionTemporal)objective.getExpression()).getOperand2();

		if (!(condition instanceof ExpressionTemporal) ||
		    !(((ExpressionTemporal)condition).getOperator() == ExpressionTemporal.P_F)) {
			throw new PrismNotSupportedException("Only reachability condition for conditional expected reward");
		}
		if (((ExpressionTemporal)condition).hasBounds()) {
			throw new PrismNotSupportedException("Only simple reachability condition for conditional expected reward, no bounds");
		}
		Expression expressionG = ((ExpressionTemporal)condition).getOperand2();

		JDDNode F = mc.checkExpressionDD(expressionF, model.getReach().copy());
		JDDNode G = mc.checkExpressionDD(expressionG, model.getReach().copy());

		// corner cases
		if (JDD.AreIntersecting(sInit, F)) {
			// (1) s_init in F  =>   accumulated reward until F = 0  =>  result = 0
			prism.getMainLog().println("s_init in F => result is 0");
			JDD.Deref(F, G);
			JDD.Deref(sInit);
			return new StateValuesMTBDD(JDD.Constant(0), model);
		}
		if (JDD.AreIntersecting(sInit, G)) {
			// (2) s_init in G  =>   Pr(F G)=1 for *all* schedulers => rely on unconditional expectations
			prism.getMainLog().println("s_init in G => compute unconditional " + objective);
			JDD.Deref(F, G);
			return mc.checkExpression(objective, sInit);
		}

		if (!F.equals(G)) {
			JDD.Deref(F, G);
			JDD.Deref(sInit);
			throw new PrismNotSupportedException("Currently, F!=G not supported for conditional MDP expectations");
		}

		JDDNode goal = F.copy();  // F = G = goal
		goalLabel = model.addUniqueLabelDD("goal", goal.copy());
		JDD.Deref(F, G);

		mc.getLog().println("\nDetermine states that can reach goal...");
		JDDNode canReachGoal = canReachGoal(model.getTransReln(), goal, model.getAllDDRowVars(), model.getAllDDColVars());
		JDDNode cantReachGoal = JDD.And(model.getReach().copy(), JDD.And(JDD.Not(goal.copy()), JDD.Not(canReachGoal.copy())));

		mc.getLog().println("States that can reach goal = " + JDD.GetNumMintermsString(canReachGoal, model.getNumDDRowVars()) + ", can't reach goal = " + JDD.GetNumMintermsString(cantReachGoal, model.getNumDDRowVars()));

		if (JDD.IsContainedIn(sInit, cantReachGoal)) {
			// corner case (3): goal can not be reached
			throw new PrismNotSupportedException("goal can not be reached");
		}

		failLabel = model.addUniqueLabelDD("fail", cantReachGoal.copy());

		mc.getLog().println("\nComputing goal-fail quotient...");
		ArrayList<JDDNode> equivalentClasses = new ArrayList<JDDNode>();
		equivalentClasses.add(goal.copy());  // -> goal state
		equivalentClasses.add(cantReachGoal.copy());  // -> fail state
		MDPQuotient quotient = MDPEquiv.transform(mc, model, equivalentClasses, model.getReach().copy());
		NondetModel quotientModel = quotient.getTransformedModel();

		if (verbose) {
			try {
				quotientModel.exportToFile(Prism.EXPORT_DOT, true, new File("goal_fail_quotient.dot"));
				quotientModel.exportStateRewardsToFile(Prism.EXPORT_PLAIN, new File("goal_fail_quotient.srew"));
				quotientModel.exportTransRewardsToFile(Prism.EXPORT_PLAIN, true, new File("goal_fail_quotient.trew"));
			} catch (FileNotFoundException e) {
			}
		}

		JDDNode goalInQuotient = quotient.mapStateSetToQuotient(goal.copy());
		if (SanityJDD.enabled) {
			SanityJDD.check(JDD.isSingleton(goalInQuotient, quotientModel.getAllDDRowVars()), "Not a single goal state");
		}

		ComputedValues computedValues = preprocessing(quotientModel, goal, min);
		if (computedValues.result != null) {
			JDD.Deref(goal);
			JDD.Deref(goalInQuotient);
			JDD.Deref(canReachGoal);
			JDD.Deref(cantReachGoal);
			JDD.Deref(sInit);
			StateValues result = quotient.projectToOriginalModel(computedValues.result);
			computedValues.result = null;  // remove result so it is not cleared, was cleared during projection

			quotient.clear();
			computedValues.clear();

			return result;
		}

		computedValues.thetaSaturated = computeThetaSaturated(computedValues);

		if (verbose) {
			mc.getLog().println();
			StateValuesMTBDD.print(mc.getLog(), computedValues.pmax.copy(), computedValues.cleanedUp.getTransformedModel(), "Pmax=?[ F goal ] for saturated strategy");
			mc.getLog().println();
			StateValuesMTBDD.print(mc.getLog(), computedValues.thetaSaturated.getJDDNode().copy(), computedValues.cleanedUp.getTransformedModel(), "Emax=?[ F goal ] for saturated strategy (alternate semantics)");
			mc.getLog().println();
		}

		phase1.stop();

		StateValuesMTBDD result = null;

		if (computedValues.allActMax) {
			// ActMax = Act, can compute CEmax directly (Remark 6.41)
			// CE^max_sinit = E_sinit / pmax_sinit,
			// as all schedulers are in SchedMax, so the scheduler optimizing
			// for the expected reward is the best one
			mc.getLog().println("ActMax = Act, we can compute the conditional expected reward directly...");
			result = new StateValuesMTBDD(JDD.Apply(JDD.DIVIDE, computedValues.thetaSaturated.getJDDNode().copy(), computedValues.pmax.copy()),
			                              model);
			result.filter(model.getStart());

			JDD.Deref(goal, canReachGoal, cantReachGoal, goalInQuotient, sInit);
			quotient.clear();
			computedValues.clear();

			return result;
		}

		mc.getLog().println("\nConverting to explicit MDP...");
		StopWatch timerConvertToExplicit = new StopWatch(mc.getLog());
		timerConvertToExplicit.start("conversion to explicit MDP");
		MDPSparse explicitMDP = new MDPSparse(computedValues.cleanedUp.getTransformedModel(), mc.getLog());
		MDPRewardsSimple explicitTransRew = new MDPRewardsSimple(JDD.ZERO, computedValues.cleanedUp.getTransformedModel().getTransRewards(0), computedValues.cleanedUp.getTransformedModel());

		mc.getLog().println("MDP for threshold/CEmax computation: " + explicitMDP.infoString());
		// mc.getLog().println("  MDP has " + countMDSchedulers(explicitMDP) + " memoryless deterministic schedulers.");

		if (verbose) {
			List<explicit.graphviz.Decorator> decorators = new ArrayList<>();
			decorators.add(new explicit.graphviz.MarkStateSetDecorator(explicitMDP.getLabelStates(goalLabel), "#30cc30"));
			decorators.add(new explicit.graphviz.MarkStateSetDecorator(explicitMDP.getLabelStates(failLabel), "#cc3030"));
			decorators.add(new explicit.graphviz.ShowRewardDecorator(explicitTransRew));
			explicitMDP.exportToDotFile(new PrismFileLog("explicit-mdp.dot"), decorators);
			mc.getLog().println("explicitTransRew = \n" + explicitTransRew.toString());
		}

		StateValuesDV svPmaxDV = new StateValuesDV(computedValues.pmax, computedValues.cleanedUp.getTransformedModel());
		explicit.StateValues svPmaxExplicit = svPmaxDV.convertToExplicitStateValues(explicitMDP);
		if (verbose) {
			mc.getLog().println("pmax:");
			svPmaxExplicit.print(mc.getLog());
		}

		StateValuesDV svEDV = new StateValuesDV(computedValues.thetaSaturated.getJDDNode(), computedValues.cleanedUp.getTransformedModel());
		explicit.StateValues svEExplicit = svEDV.convertToExplicitStateValues(explicitMDP);
		if (verbose) {
			mc.getLog().println("E:");
			svEExplicit.print(mc.getLog());
		}

		svPmaxDV.clear();
		svEDV.clear();
		timerConvertToExplicit.stop();

		Integer rho;

		if (saturationPointMethod == SaturationPointMethod.DIRECT) {
			rho = computeSaturationPointDirect(explicitMDP, computedValues.consts.CEmaxUpperBound, explicitTransRew, svPmaxExplicit, svEExplicit);
		} else {
			rho = computedValues.consts.pSaturation;
		}

		int override = mc.getSettings().getInteger(PrismSettings.CONDITIONAL_EXP_MDP_PSATURATION);
		if (override >= 0) {
			if (rho != null) {
				mc.getLog().printWarning("Overriding saturation point from " + rho + " to " + override + ".");
			} else {
				mc.getLog().printWarning("Using manual saturation point " + override + ".");
			}
			rho = override;
		}

		ThresholdAlgorithm info;
		if (computedValues.isAcyclic) {
			info = new ThresholdAlgorithmAcyclic(explicitMDP, explicitTransRew, svPmaxExplicit, svEExplicit);
		} else {
			if (rho == null) {
				throw new PrismException("Need saturation point, either enable computation method (-cexpmdplambda) or provide one manually (-cexpmdpsaturation)");
			}
			info = new ThresholdAlgorithmGeneralCase(explicitMDP, explicitTransRew, rho, svPmaxExplicit, svEExplicit);
		}

		OpRelOpBound relOpInfo = objective.getRelopBoundInfo(mc.getConstantValues());
		if (!relOpInfo.isNumeric()) {
			double threshold = relOpInfo.getBound();
			boolean lower = relOpInfo.getRelOp().isLowerBound();
			boolean strict = relOpInfo.getRelOp().isStrict();

			String thres = relOpInfo.getRelOp().toString() + " " + threshold;
			boolean actualStrict = lower ? strict : !strict;

			if (lower) {
				mc.getLog().println("Checking CEmax threshold " + thres);
			} else {
				mc.getLog().println("For checking threshold " + thres + ", we actually check CEmax threshold " + (strict ? ">= " : "> ") + threshold + " and negate.");
			}

			StopWatch timerThreshold = new StopWatch(mc.getLog());
			timerThreshold.start("threshold computation");

			OptimalChoicesComputed<OptimalChoice> witness = 
					evaluateThresholdLower(info,
							threshold,
							0.0,
							computedValues.consts.CEmaxUpperBound,
							actualStrict,
							rho);

			timerThreshold.stop();

			mc.getLog().println("\n============== Result ====================\n");

			boolean rv;
			if (witness == null) {
				if (lower) {
					mc.getLog().println("Could not find witness for CEmax " + thres + ".");
					rv = false;
				} else {
					mc.getLog().println("Could not find counter-example for CEmax " + thres + ".");
					rv = true;
				}
			} else {
				int exp_sInit = explicitMDP.getFirstInitialState();
				double CEwitness = witness.getFor(0,exp_sInit).getCE();

				if (lower) {
					mc.getLog().println("Found witness for CEmax " + thres  + ", with CE = " + CEwitness + ".");
					rv = true;
				} else {
					mc.getLog().println("Found counter-example for CEmax " + thres +", with CE = " + CEwitness + ".");
					rv = false;
				}

				if (mc.getSettings().getBoolean(PrismSettings.CONDITIONAL_EXP_MDP_SHOW_SCHEDULER)) {
					mc.getLog().println();
					mc.getLog().println(constructScheduler(witness));
				}
			}

			if (rv) {
				result = new StateValuesMTBDD(JDD.ITE(sInit.copy(), JDD.Constant(1), JDD.Constant(0)), model);
			} else {
				result = new StateValuesMTBDD(JDD.Constant(0), model);
			}
		} else {
			// quantitative computation of CEmax
			boolean failed = false;
			try {
				double CEmax;
				switch (mc.getSettings().getString(PrismSettings.CONDITIONAL_EXP_MDP_METHOD)) {
				case "Scheduler Improvement":
					CEmax = computeCEMax(info, computedValues.consts.CEmaxUpperBound);
					break;
				case "Sequential":
					CEmax = computeCEMaxSequential(info);
					break;
				default:
					throw new PrismException("Unknown CONDITONAL_EXP_MDP_METHOD");
				}

				if (mc.getSettings().getBoolean(PrismSettings.CONDITIONAL_EXP_MDP_SHOW_SCHEDULER)) {
					mc.getLog().println("\n" + constructScheduler(info.getOpt()));
				}

				mc.getLog().println("============== Result ====================\n");
				mc.getLog().println("We have computed the CEmax value: CEmax = " + CEmax + ", upper bound was " + computedValues.consts.CEmaxUpperBound);

				if (computedValues.consts.CEmaxUpperBound < CEmax && !doublesAreClose(computedValues.consts.CEmaxUpperBound, CEmax, 1E-6, false)) {
					mc.getLog().printWarning("The computed CEmax value is larger than the upper bound. If the values are close, this is most likely due to numerical imprecision in the computation of the upper bound");
				}

				result = new StateValuesMTBDD(JDD.ITE(model.getStart().copy(), JDD.Constant(CEmax), JDD.Constant(0)), model);
			} catch (Exception e) {
				failed = true;
				e.printStackTrace();
			}

			if (failed) {
				// return CEMaxUpperbound as (provisional) best guess at result
				mc.getLog().printWarning("Result for conditional expectation is not necessarily correct, result is best guess at an upper bound!");
				result = new StateValuesMTBDD(JDD.ITE(model.getStart().copy(), JDD.Constant(computedValues.consts.CEmaxUpperBound), JDD.Constant(0)), model);
			}
		}

		JDD.Deref(goal, canReachGoal, cantReachGoal, goalInQuotient, sInit);
		quotient.clear();
		computedValues.clear();

		return result;
	}

	private ComputedValues preprocessing(NondetModel model, JDDNode goal, boolean min) throws PrismException
	{
		if (min) {
			return null;
		}

		NondetModelChecker mc = this.mc.createNewModelChecker(prism, model, this.mc.getPropertiesFile());

		JDDNode statesWithPosRew = JDD.GreaterThan(model.getStateRewards(0).copy(), 0);
		JDDNode transWithPosRew = JDD.GreaterThan(model.getTransRewards(0).copy(), 0);

		// ----- first, search for positive end components

		mc.getLog().println("\n\nPrecomputation: Searching for positive end components...");

		ECComputer ecs = ECComputer.createECComputer(mc, model);
		ecs.computeMECStates();
		int i = 0;
		boolean foundPosEC = false;
		for (JDDNode ec : ecs.getMECStates()) {
			i++;
			foundPosEC = isPosEC(ecs, ec, statesWithPosRew, transWithPosRew);
			if (foundPosEC) {
				break;
			}
		}

		mc.getLog().print("Checked " + i + " of " + ecs.getMECStates().size() + " maximal end components, ");

		if (foundPosEC) {
			mc.getLog().println("found at least one that is positive.");
			// max conditional expected reward is infinite
			JDD.Deref(statesWithPosRew, transWithPosRew);
			for (JDDNode ec : ecs.getMECStates()) {
				JDD.Deref(ec);
			}
			JDDNode result = JDD.ITE(model.getStart().copy(), JDD.PlusInfinity(), JDD.Constant(0));
			return new ComputedValues(new StateValuesMTBDD(result, model));
		}
		mc.getLog().println("none are positive.");


		// ---- second, check for critical schedulers 

		// remember states in MECs
		JDDNode inECs = JDD.Constant(0);
		for (JDDNode ec : ecs.getMECStates()) {
			inECs = JDD.Or(inECs, ec.copy());
		}

		mc.getLog().println("\n\nPrecomputation: Searching for critical schedulers in MEC quotient MDP...");
		// note: quotienting derefs the ECs 
		MDPQuotient mecQuotient = MDPEquiv.transform(mc, model, ecs.getMECStates(), model.getReach().copy());
		NondetModel mecQuotientModel = mecQuotient.getTransformedModel();
		mc.getLog().println("MEC quotient MDP: " + mecQuotientModel.infoString() + ".");

		if (verbose) {
			try {
				mecQuotientModel.exportToFile(Prism.EXPORT_DOT, true, new File("mec_quotient.dot"));
				mecQuotientModel.exportStateRewardsToFile(Prism.EXPORT_PLAIN, new File("mec_quotient.srew"));
				mecQuotientModel.exportTransRewardsToFile(Prism.EXPORT_PLAIN, true, new File("mec_quotient.trew"));
			} catch (FileNotFoundException e) {
			}
		}

		if (hasCriticalScheduler(mecQuotientModel, statesWithPosRew, transWithPosRew)) {
			// max conditional expected reward is infinite
			JDDNode result = JDD.ITE(model.getStart().copy(), JDD.PlusInfinity(), JDD.Constant(0));

			mecQuotient.clear();
			JDD.Deref(statesWithPosRew, transWithPosRew, inECs);

			return new ComputedValues(new StateValuesMTBDD(result, model));
		}

		mc.getLog().println();

		JDDNode failState = mecQuotientModel.getLabelDD(failLabel).copy();
		failState = JDD.And(failState, mecQuotientModel.getReach().copy());
		JDDNode tauStates = JDD.And(inECs.copy(), JDD.Not(mecQuotientModel.getLabelDD(goalLabel).copy()));
		tauStates = JDD.And(tauStates, mecQuotientModel.getReach().copy());
		tauStates = JDD.Or(tauStates, failState.copy());

		if (tauStates.equals(JDD.ZERO)) {
			ExpressionReward e = new ExpressionReward();
			e.setMinMax(min ? MinMax.min() : MinMax.max());
			e.setRelOp(RelOp.COMPUTE_VALUES);
			e.setExpression(new ExpressionTemporal(ExpressionTemporal.P_F, null, new ExpressionLabel(goalLabel)));
			mc.getLog().println("'fail' not reachable  =>  Pmin[F goal]=1  =>  compute unconditional expections : " + e);

			NondetModelChecker mcMeqQuotient = mc.createNewModelChecker(prism, mecQuotientModel, mc.getPropertiesFile());
			JDDNode result = mcMeqQuotient.checkExpressionDD(e, mecQuotientModel.getReach().copy());
			StateValuesMTBDD r = new StateValuesMTBDD(result, mecQuotientModel);
			r = mecQuotient.projectToOriginalModel(r).convertToStateValuesMTBDD();

			mecQuotient.clear();
			JDD.Deref(statesWithPosRew, transWithPosRew, failState, tauStates, inECs);

			return new ComputedValues(r);
		}

		MDPWithFail cleanedUp = MDPAddTauToFail.transform(mc, mecQuotientModel, tauStates.copy(), failState.copy());
		NondetModel cleanedModel = cleanedUp.getTransformedModel();
		// ensure that the fail state in the cleaned up model is labeled with failLabel
		cleanedModel.addLabelDD(failLabel, cleanedUp.getFailState());

		// move state rewards into transition rewards as well 
		JDDNode rewards = cleanedModel.getStateRewards(0).copy();
		rewards = JDD.Apply(JDD.PLUS, rewards, cleanedModel.getTransRewards(0).copy());
		rewards = JDD.Apply(JDD.TIMES, rewards, cleanedModel.getTrans01().copy());
		cleanedModel.resetStateRewards(0, JDD.Constant(0));
		cleanedModel.resetTransRewards(0, rewards);

		if (verbose) {
			try {
				cleanedModel.exportToFile(Prism.EXPORT_DOT, true, new File("cleaned.dot"));
				cleanedModel.exportStateRewardsToFile(Prism.EXPORT_PLAIN, new File("cleaned.srew"));
				cleanedModel.exportTransRewardsToFile(Prism.EXPORT_PLAIN, true, new File("cleaned.trew"));
			} catch (FileNotFoundException e) {
			}
		}

		if (isLTS(cleanedModel) &&
			saturationPointMethod != SaturationPointMethod.SKIP &&
			saturationPointMethod != SaturationPointMethod.DIRECT) {
			// only do special LTS handling for inefficient saturation point methods
			JDDNode result = computeForLTS(cleanedModel);
			StateValuesMTBDD r = new StateValuesMTBDD(result, cleanedModel);

			mecQuotient.clear();
			cleanedUp.clear();
			JDD.Deref(statesWithPosRew, transWithPosRew, failState, tauStates, inECs);

			return new ComputedValues(r);
		}

		double upperBound = computeMaxUpperBound(cleanedModel);
		mc.getLog().println("Upper bound = " + upperBound);

		boolean doSpecialAcyclic = false;
		boolean isAcyclic = false;
		if (doSpecialAcyclic) {
			isAcyclic = checkIsAcyclic(cleanedModel);
		}

		// compute saturation constants and pMax, actMax etc
		// is MDP is acyclic, don't estimate the saturation constants
		ComputedValues computedValues = computeSaturationConstants(cleanedModel, upperBound, 2, isAcyclic);

		if (!isAcyclic) {
			mc.getLog().println("\nSaturation constants:\n" + computedValues.consts.toString(true));
		}
		computedValues.isAcyclic = isAcyclic;

//		For evaluating effect of k:
//
//		StringBuffer r = new StringBuffer();
//		for (int k=2; k<10; k++) {
//			Constants c = computeSaturationConstants(cleanedModel, k);
//			r.append("k = " + k + ": " + c + "\n");
//		}
//
//		mc.getLog().printSeparator();
//		mc.getLog().println(r);
//		mc.getLog().printSeparator();

		computedValues.mecQuotient = mecQuotient;
		computedValues.cleanedUp = cleanedUp;
		JDD.Deref(statesWithPosRew, transWithPosRew, failState, tauStates, inECs);

		return computedValues;
	}

	private boolean checkIsAcyclic(NondetModel mdp) throws PrismException
	{
		StopWatch timer = new StopWatch(mc.getLog());

		timer.start("cycle check");
		mc.getLog().println("Performing cycle check on cleaned-up MDP...");
		SCCComputer sccs = SCCComputer.createSCCComputer(mc, mdp);
		JDDNode specialStates = mdp.getLabelDD(goalLabel).copy();
		specialStates = JDD.Or(specialStates, mdp.getLabelDD(failLabel).copy());
		StateValuesMTBDD.print(mc.getLog(), specialStates.copy(), mdp, "SpecialStates");
		sccs.computeSCCs();

		boolean hadSCC = false;
		for (JDDNode scc : sccs.getSCCs()) {
			JDD.Deref(scc);
			StateValuesMTBDD.print(mc.getLog(), scc.copy(), mdp, "SCC");
			if (!JDD.IsContainedIn(scc, specialStates)) {
				// we only care for cycles in the non-special state space
				hadSCC = true;
			}
		}
		JDD.Deref(sccs.getNotInSCCs());
		JDD.Deref(specialStates);
		timer.stop(hadSCC ? " (MDP has cycles)" : " (MDP is acyclic)");

		return !hadSCC;
	}

	private boolean isPosEC(ECComputer ecs, JDDNode ec, JDDNode statesWithPosRew, JDDNode transWithPosRew) {
		boolean hasStateRewards = !(statesWithPosRew.equals(JDD.ZERO));
		boolean hasTransRewards = !(transWithPosRew.equals(JDD.ZERO));

		if (hasStateRewards && JDD.AreIntersecting(ec, statesWithPosRew)) {
			return true;
		}

		if (hasTransRewards) {
			// obtain EC transitions
			JDDNode ecTransitions = ecs.getStableTransitions(ec.copy());
			if (JDD.AreIntersecting(ecTransitions, transWithPosRew)) {
				JDD.Deref(ecTransitions);
				return true;
			}
			JDD.Deref(ecTransitions);
		}

		return false;
	}

	private boolean hasCriticalScheduler(NondetModel mecQuotientModel, JDDNode statesWithPosRew, JDDNode transWithPosRew) throws PrismException
	{
		JDDNode goalInQuotient = mecQuotientModel.getLabelDD(goalLabel).copy();
		
		if (verbose) StateValuesMTBDD.print(mc.getLog(), goalInQuotient.copy(), mecQuotientModel, "goalInQuotient");

		// Find the set of states of M\goal, i.e., the MDP where a scheduler can
		// remain indefinitely without reaching goal. Note that, as every state but
		// fail can reach goal later on by assumption, a positive cycle in this
		// are provides a critical scheduler.
		mc.getLog().println("Computing Pmin=0[ F goal ] states for M\\goal");
		JDDNode stableNotGoal = PrismMTBDD.Prob0E(mecQuotientModel.getTrans01(),
		                                          mecQuotientModel.getReach(),
		                                          mecQuotientModel.getNondetMask(),
		                                          mecQuotientModel.getAllDDRowVars(),
		                                          mecQuotientModel.getAllDDColVars(),
		                                          mecQuotientModel.getAllDDNondetVars(),
		                                          mecQuotientModel.getReach(),
		                                          goalInQuotient);

		if (!JDD.IsContainedIn(mecQuotientModel.getStart(), stableNotGoal)) {
			mc.getLog().println("Pmin[ F goal ] > 0 for the state of interest => there are no critical schedulers");
			JDD.Deref(goalInQuotient);
			JDD.Deref(stableNotGoal);
			return false;
		}

		if (verbose) StateValuesMTBDD.print(mc.getLog(), stableNotGoal.copy(), mecQuotientModel, "stableNotGoal");

		ECComputer ecMecQuotient = ECComputer.createECComputer(mc, mecQuotientModel);
		JDDNode transInStable = ecMecQuotient.getStableTransitions(stableNotGoal.copy());
		JDDNode transReln = JDD.ThereExists(transInStable.copy(), mecQuotientModel.getAllDDNondetVars());

		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transReln.copy(), "transReln M\\goal");
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transInStable.copy(), "transInStable M\\goal");

		// ensure that we only keep states that are reachable from s_init
		mc.getLog().println("Compute reachability for M\\goal");
		JDDNode reach = PrismMTBDD.Reachability(transReln, mecQuotientModel.getAllDDRowVars(), mecQuotientModel.getAllDDColVars(), mecQuotientModel.getStart());
		transReln = JDD.And(reach, transReln);

		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transReln.copy(), "transReln M\\goal (reachable)");

		mc.getLog().println(" Searching for cycle in MEC quotient MDP M\\goal (computing SCCs)...");
		SCCComputer sccs = SCCComputer.createSCCComputer(prism, stableNotGoal, transReln, mecQuotientModel.getAllDDRowVars(), mecQuotientModel.getAllDDColVars());
		sccs.computeSCCs();
		JDD.Deref(sccs.getNotInSCCs());
		boolean foundPositiveCycle = false;
		for (JDDNode scc : sccs.getSCCs()) {
			if (foundPositiveCycle) {
				JDD.Deref(scc);
				continue;
			}

			if (extraVerbose) StateValuesMTBDD.print(mc.getLog(), scc.copy(), mecQuotientModel, "scc");
			if (extraVerbose) JDD.PrintMinterms(mc.getLog(), scc.copy(), "scc");

			if (JDD.AreIntersecting(scc, statesWithPosRew)) {
				foundPositiveCycle = true;
				JDD.Deref(scc);
				continue;
			}

			JDDNode sccTrans = JDD.And(scc.copy(), transInStable.copy());
			sccTrans = JDD.And(sccTrans, JDD.PermuteVariables(scc.copy(), mecQuotientModel.getAllDDRowVars(), mecQuotientModel.getAllDDColVars()));
			if (extraVerbose) JDD.PrintMinterms(mc.getLog(), sccTrans.copy(), "sccTrans");
			if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transWithPosRew.copy(), "transWithPosRew");
			if (JDD.AreIntersecting(sccTrans, transWithPosRew)) {
				foundPositiveCycle = true;
			}
			JDD.Deref(sccTrans, scc);
		}

		JDD.Deref(transReln);
		JDD.Deref(stableNotGoal);
		JDD.Deref(goalInQuotient);
		JDD.Deref(transInStable);

		if (foundPositiveCycle) {
			mc.getLog().println("Found a positive cycle => There exists a critical scheduler.");
		} else {
			mc.getLog().println("Found no positive cycle => There is no critical scheduler.");
		}
		return foundPositiveCycle;
	}

	private double computeMaxUpperBound(NondetModel model) throws PrismException
	{
		mc.getLog().println("\nComputing upper bound for CEmax...");
		StopWatch timerBuild = new StopWatch(mc.getLog());
		timerBuild.start("construction of the MDP for upper bound estimation");

		int R = computeR(model);
		mc.getLog().println("Construction product model for R = " + R + ".");

		JDDNode rewards = model.getStateRewards(0).copy();
		rewards = JDD.Apply(JDD.PLUS, rewards, model.getTransRewards(0).copy());
		rewards = JDD.Apply(JDD.TIMES, rewards, model.getTrans01().copy());
		
		if (verbose) {
			mc.getLog().println("MDP M':");
			model.printTransInfo(mc.getLog(), true);
		}

		RewardCounterProduct<NondetModel> product = RewardCounterProduct.generate(prism, model, rewards, R+1, model.getStart().copy());
		NondetModel productModel = product.getProductModel();

		// fix up rewards:
		// - first, transform state rewards to trans rewards
		// rew[ (s,r) -alpha -> (s',r') ] =
		//    0                       for r' < R+1 and s' != goal
		//    rew[s -alpha-> s']      for r  = R+1
		//    rew[s -alpha-> s'] + r  for r <= R   and r' = R+1
		//                            or s != goal and s = goal

		JDDNode transRew = productModel.getTransRewards(0).copy();
		transRew = JDD.Apply(JDD.PLUS, transRew, productModel.getStateRewards(0).copy());
		productModel.resetStateRewards(0, JDD.Constant(0));
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transRew.copy(), "transRew");

		JDDNode statesRplus1 = product.getStatesWithAccumulatedReward(R+1);
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), statesRplus1.copy(), "statesRplus1");

		IntegerBound lessThanRplus1 = new IntegerBound(0, false, R+1, true);
		JDDNode statesLessThanRplus1 = product.getStatesWithAccumulatedRewardInBound(lessThanRplus1);
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), statesLessThanRplus1.copy(), "statesLessThanRplus1");

		JDDNode goal = productModel.getLabelDD(goalLabel).copy();
		goal = JDD.And(goal, productModel.getReach().copy());
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), goal.copy(), "goal");

		JDDNode targetIsFirstMode = JDD.And(statesLessThanRplus1.copy(), JDD.Not(goal.copy()));
		
		// r' < R+1 & s' != goal  => reward 0
		transRew = JDD.ITE(JDD.PermuteVariables(targetIsFirstMode, productModel.getAllDDRowVars(), productModel.getAllDDColVars()),
		                   JDD.Constant(0),
		                   transRew);
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transRew.copy(), "transRew(2)");

		JDDNode transRewModeSwitch = JDD.Constant(0);
		for (int i = 0; i <= R; i++) {
			transRewModeSwitch = JDD.Apply(JDD.PLUS, transRewModeSwitch,
			                               JDD.ITE(product.getStatesWithAccumulatedReward(i), JDD.Constant(i), JDD.Constant(0)));
		}
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transRewModeSwitch.copy(), "transRewModeSwitch");
		
		JDDNode transModeSwitchRtoRplus1 =
			JDD.And(statesLessThanRplus1.copy(),
			        JDD.PermuteVariables(statesRplus1.copy(), productModel.getAllDDRowVars(), productModel.getAllDDColVars()));
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transModeSwitchRtoRplus1.copy(), "transModeSwitchRtoRplus1");
		
		JDDNode transModeSwitchToGoal = JDD.And(JDD.Not(goal.copy()),
		                                        JDD.PermuteVariables(goal.copy(), productModel.getAllDDRowVars(), productModel.getAllDDColVars()));
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transModeSwitchToGoal.copy(), "transModeSwitchToGoal");

		JDDNode transModeSwitch = JDD.Or(transModeSwitchRtoRplus1, transModeSwitchToGoal);
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transModeSwitch.copy(), "transModeSwitch");
		
		transRewModeSwitch = JDD.Times(transRewModeSwitch, transModeSwitch);
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transRewModeSwitch.copy(), "transRewModeSwitch");

		transRew = JDD.Apply(JDD.PLUS, transRew, transRewModeSwitch);
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transRew.copy(), "transRew (3)");

		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), productModel.getTrans01().copy(), "productModel.trans01");
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), productModel.getTrans().copy(), "productModel.trans");

		transRew = JDD.Times(transRew, productModel.getTrans01().copy());
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), transRew.copy(), "transRew (4)");

		productModel.resetTransRewards(0, transRew);

		if (verbose) {
			try {
				productModel.exportToFile(Prism.EXPORT_DOT, true, new File("productR.dot"));
				productModel.exportStateRewardsToFile(Prism.EXPORT_PLAIN, new File("productR.srew"));
				productModel.exportTransRewardsToFile(Prism.EXPORT_PLAIN, true, new File("productR.trew"));
			} catch (FileNotFoundException e) {
			}
		}

		JDDNode resetStates = productModel.getLabelDD(failLabel).copy();
		resetStates = JDD.And(resetStates, productModel.getReach().copy());
		JDDNode resetTarget = productModel.getStart().copy(); 

		MDPResetTransformation resetTransform = new MDPResetTransformation(productModel, resetStates.copy(), resetTarget.copy(), true, mc.getLog());
		NondetModel resetModel = productModel.getTransformed(resetTransform);

		if (verbose) {
			try {
				mc.getLog().println("productR-reset: " + resetModel.infoString());
				resetModel.exportToFile(Prism.EXPORT_DOT, true, new File("productR-reset.dot"));
				resetModel.exportStateRewardsToFile(Prism.EXPORT_PLAIN, new File("productR-reset.srew"));
				resetModel.exportTransRewardsToFile(Prism.EXPORT_PLAIN, true, new File("productR-reset.trew"));
			} catch (FileNotFoundException e) {
			}
		}

		MDPQuotient resetQuotient = MDPEquiv.mecQuotient(mc, resetModel, resetModel.getReach().copy(), resetModel.getReach().copy());
		NondetModel resetQuotientModel = resetQuotient.getTransformedModel();

		if (verbose) {
			try {
				mc.getLog().println("productR-reset-q: " + resetQuotientModel.infoString());
				resetQuotientModel.exportToFile(Prism.EXPORT_DOT, true, new File("productR-reset-q.dot"));
				resetQuotientModel.exportStateRewardsToFile(Prism.EXPORT_PLAIN, new File("productR-reset-q.srew"));
				resetQuotientModel.exportTransRewardsToFile(Prism.EXPORT_PLAIN, true, new File("productR-reset-q.trew"));
			} catch (FileNotFoundException e) {
			}
		}

		timerBuild.stop();

		StopWatch timerCompute = new StopWatch(mc.getLog());
		timerCompute.start("computing upper bound for CEmax in product MDP");
		ExpressionReward e = new ExpressionReward();
		e.setMinMax(MinMax.max());
		e.setRelOp(RelOp.COMPUTE_VALUES);
		e.setExpression(new ExpressionTemporal(ExpressionTemporal.P_F, null, new ExpressionLabel(goalLabel)));
		mc.getLog().println("Compute unconditional expected reward in MDP N: " + e);

		NondetModelChecker mcResetQuotient = mc.createNewModelChecker(prism, resetQuotientModel, mc.getPropertiesFile());
		JDDNode result = mcResetQuotient.checkExpressionDD(e, resetQuotientModel.getStart().copy());
		StateValuesMTBDD r = new StateValuesMTBDD(result, resetQuotientModel);
		r = resetQuotient.projectToOriginalModel(r).convertToStateValuesMTBDD();

		r.filter(resetTarget);
		StateValuesMTBDD r2 = r.sumOverDDVars(resetModel.getAllDDRowVars(), resetModel).convertToStateValuesMTBDD();
		SanityJDD.check(r2.getJDDNode().isConstant(), "Value for start state should be constant");
		double value = r2.getJDDNode().getValue();
		r.clear();
		r2.clear();

		timerCompute.stop();

		resetQuotient.clear();
		resetModel.clear();
		resetTransform.clear();
		product.clear();

		JDD.Deref(statesLessThanRplus1, statesRplus1, goal, resetTarget, resetStates);

		return value;
	}

	private JDDNode canReachGoal(JDDNode transRel, JDDNode goal, JDDVars ddRowVars, JDDVars ddColVars) {
		JDDNode T = goal.copy();
		JDDNode oldT = null;
		int iterations = 0;
		do {
			if (oldT != null) JDD.Deref(oldT);
			oldT = T.copy();
			T = JDD.And(transRel.copy(), JDD.PermuteVariables(T, ddRowVars, ddColVars));
			T = JDD.ThereExists(T, ddColVars);
			iterations++;
		} while (!T.equals(oldT));
		JDD.Deref(oldT);

		mc.getLog().println("Backward reachability for \"goal\": " + iterations + " iterations.");

		return T;
	}

	private boolean isLTS(NondetModel model)
	{
		// TODO: do it graph-based

		double pmin = JDD.FindMinPositive(model.getTrans());
		return (pmin == 1.0);
	}

	private static class LTSNotFailTransformation extends NondetModelTransformation {

		private String failLabel;

		public LTSNotFailTransformation(NondetModel model, String failLabel)
		{
			super(model);

			this.failLabel = failLabel;
		}

		@Override
		public int getExtraStateVariableCount()
		{
			return 0;
		}

		@Override
		public int getExtraActionVariableCount()
		{
			return 0;
		}

		@Override
		public JDDNode getTransformedTrans() throws PrismException
		{
			// we are in the LTS situation
			//  * all choices have exactly one successor state
			//  * due to (A1) and (A2), all states (except fail)
			//    have a path to goal
			//  * we remove the fail state and all incoming choices
			JDDNode transNotFail = originalModel.getTrans().copy();
			JDDNode failRow = originalModel.getLabelDD(failLabel).copy();
			JDDNode failCol = JDD.PermuteVariables(failRow.copy(), originalModel.getAllDDRowVars(), originalModel.getAllDDColVars());

			transNotFail = JDD.Times(transNotFail, JDD.Not(failCol));
			transNotFail = JDD.Times(transNotFail, JDD.Not(failRow));

			return transNotFail;
		}

		@Override
		public JDDNode getTransformedStart() throws PrismException
		{
			return originalModel.getStart().copy();
		}

		@Override
		public JDDNode getReachableStates()
		{
			return JDD.And(originalModel.getReach().copy(), JDD.Not(originalModel.getLabelDD(failLabel).copy()));
		}
	}

	private JDDNode computeForLTS(NondetModel model) throws PrismException
	{
		// we are in the LTS situation
		//  * all choices have exactly one successor state
		//  * due to (A1) and (A2), all states (except fail)
		//    have a path to goal
		//  * we remove the fail state and all incoming choices

		mc.getLog().println("Cleaned-up MDP is actually an LTS (all probabilities = 1), computing expected reward directly...");

		LTSNotFailTransformation transform = new LTSNotFailTransformation(model, failLabel);
		NondetModel transformed = model.getTransformed(transform);

		NondetModelChecker mcTransformed = (NondetModelChecker) mc.createModelChecker(transformed);
		ExpressionTemporal eTemp = new ExpressionTemporal(ExpressionTemporal.P_F, null, new ExpressionLabel(goalLabel));
		ExpressionReward e = new ExpressionReward(eTemp, MinMax.max(), "=", null);
		JDDNode result = mcTransformed.checkExpressionDD(e, transformed.getStart().copy());

		transformed.clear();
		transform.clear();

		return JDD.ITE(model.getStart().copy(), result, JDD.Constant(0));
	}

	private int computeR(NondetModel model) throws PrismException
	{
		JDDNode stateRews = model.getStateRewards(0).copy();
		JDDNode transRews = model.getTransRewards(0).copy();

		transRews = JDD.Apply(JDD.PLUS, stateRews, transRews);
		transRews = JDD.Times(transRews, model.getReach().copy());

		JDDNode maxRews = JDD.MaxAbstract(transRews, model.getAllDDNondetVars());
		maxRews = JDD.MaxAbstract(maxRews, model.getAllDDColVars());

		JDDNode R = JDD.SumAbstract(maxRews, model.getAllDDRowVars());

		int result = common.SafeCast.toInt(R.getValue());

		JDD.Deref(R);
		return result;
	}

	private ComputedValues computeSaturationConstants(NondetModel model, double CEmaxUpperBound, int k, boolean noSaturationPoint) throws PrismException
	{
		ComputedValues computedValues = new ComputedValues();
		Constants consts = new Constants();

		if (!noSaturationPoint) {
			switch (saturationPointMethod) {
			case NAIVE:
				estimatePNNaively(consts, model);
				break;
			case REWARD_QUANTILE:
				estimatePNRewardQuantile(consts, model);
				break;
			case DIRECT:
				noSaturationPoint = true;
				break;
			case SKIP:
				noSaturationPoint = true;
				break;
			default:
				throw new PrismException("Unknown method");
			}
		}

		consts.CEmaxUpperBound = CEmaxUpperBound;
		computedValues.consts = consts;

		NondetModelChecker mc = this.mc.createNewModelChecker(prism, model, this.mc.getPropertiesFile());
		ExpressionProb e = new ExpressionProb();
		e.setMinMax(MinMax.max());
		e.setRelOp(RelOp.COMPUTE_VALUES);
		e.setExpression(new ExpressionTemporal(ExpressionTemporal.P_F, null, new ExpressionLabel(goalLabel)));
		JDDNode ddPmax = mc.checkExpressionDD(e, model.getReach().copy());

		StateValuesMTBDD pMax = new StateValuesMTBDD(ddPmax, model);
		// p = min_{s, s != fail} Pmax=?[ F goal ]  <=>  p = min_pos Pmax=?[ F goal ]
		consts.p_6_35 = JDD.FindMinPositive(pMax.getJDDNode());

		if (!noSaturationPoint) {
			int R = consts.R0;
			if (R % 2 > 0)
				R++;
			while (true) {
				double v = consts.C * R * Math.pow(consts.lambda, R/2);
				if (v < consts.p_6_35/2.0) {
					break;
				}
				R += 2;
			}
			consts.Rturning = R;
		}

		JDDNode pmaxRow = pMax.getJDDNode().copy();
		JDDNode pmaxCol = JDD.PermuteVariables(pmaxRow.copy(), model.getAllDDRowVars(), model.getAllDDColVars());

		// delta : s,alpha -> R with delta(s,alpha) = sum_t p(s,alpha,t) * pmax_t
		// Note: delta(s,alpha) <= pmax_s (apart from rounding errors)
		JDDNode delta = JDD.Times(model.getTrans().copy(), pmaxCol);
		delta = JDD.SumAbstract(delta, model.getAllDDColVars());

		JDDNode deltaPos = JDD.GreaterThan(delta.copy(), 0);

		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), deltaPos.copy(), "deltaPos");
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), delta.copy(), "delta");
		
		// diff : s,alpha -> R with diff(s,alpha) = pmax_s - delta(s,alpha)
		// Note: diff(s,alpha) <= 0  =>  pmax_s = delta(s,alpha) 
		JDDNode diff = JDD.Apply(JDD.MINUS, pmaxRow.copy(), delta.copy());
		
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), diff.copy(), "diff");

		// actMax : s,alpha -> {0,1} with actMax(s,alpha) = 1 iff delta(s,alpha) >= pmax_s
		JDDNode actMax = JDD.LessThanEquals(diff, 0.0);  // TODO: compare against epsilon? Use graph-based / LP derived set of actions?

		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), actMax.copy(), "actMax");

		actMax = JDD.And(actMax, deltaPos);
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), actMax.copy(), "actMax2");
		
		computedValues.actMax = actMax.copy();
		computedValues.pmax = pmaxRow.copy();

		// Theorem 6.38
		//   delta_not_max(s,a) = delta(s,a) for all a not in ActMax, 0 otherwise 
		JDDNode delta_not_max = JDD.ITE(actMax.copy(), JDD.Constant(0), delta.copy());
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), delta.copy(), "delta");
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), delta_not_max.copy(), "delta_not_max");
		//   diff_not_max(s,a) = pmax_s - delta_not_max(s,a)
		JDDNode diff_not_max = JDD.Apply(JDD.MINUS, pmaxRow.copy(), delta_not_max);
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), diff_not_max.copy(), "diff_not_max");
		//   set all a in ActMax to 0
		diff_not_max = JDD.ITE(actMax.copy(), JDD.Constant(0), diff_not_max);
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), diff_not_max.copy(), "diff_not_max (2)");
		//   set all choices not in Act(s) to 0
		diff_not_max = JDD.ITE(model.getNondetMask().copy(), JDD.Constant(0), diff_not_max);
		if (extraVerbose) JDD.PrintMinterms(mc.getLog(), diff_not_max.copy(), "diff_not_max (3)");
		consts.delta_6_38 = JDD.FindMinPositive(diff_not_max);

		if (consts.delta_6_38 == Double.POSITIVE_INFINITY) {
			// ActMax = Act, can compute CEmax directly (Remark 6.41)
			computedValues.allActMax = true;
		}

		
		if (!noSaturationPoint) {
			// TODO: special handling of case where lambda is 0
			// -> We know that after R0 we are in a trap with probability 1

			double pSaturation = Math.ceil(consts.CEmaxUpperBound + (consts.Rturning/consts.delta_6_38));
			long pSatInt = Math.round(pSaturation);
			if (pSatInt > Integer.MAX_VALUE) {
				throw new PrismException("pSaturation is too large, not representable as an int: " + pSaturation);
			}
			assert(pSatInt >= 0);
			consts.pSaturation = (int) pSatInt;
		}

		JDD.Deref(pmaxRow);
		JDD.Deref(delta);
		JDD.Deref(diff_not_max);
		JDD.Deref(actMax);

		pMax.clear();

		return computedValues;
	}

	/**
	 * Compute estimate for N and p_6_17.
	 * This estimate is naive:
	 * <ul>
	 *  <li>p_6_17 = minimal non-zero transition probability</li>
	 *  <li>N = number of states</li>
	 * </ul>
	 * <br>
	 * Stores results in consts
	 */
	private void estimatePNNaively(Constants consts, NondetModel model) throws PrismException
	{
		double R_ = JDD.FindMax(model.getTransRewards(0));
		int R = SafeCast.toInt(R_);

		double N_ = JDD.GetNumMinterms(model.getReach(), model.getNumDDRowVars());
		int N = SafeCast.toInt(N_);

		double p = JDD.FindMinPositive(model.getTrans());
		SanityJDD.check(model.getStateRewards(0).equals(JDD.ZERO), "Model has state rewards, can only handle transition rewards");

		double pN = Math.pow(p, N);

		consts.details_6_17 = "N = " + N + ", R = " + R + ", p(6.17) = " + p + ", p^N = " + pN + ", 1-p^N = " + (1.0-pN);

		computeLambda(consts, 2, pN, N*R);
	}

	/**
	 * Compute estimate for N and p_6_17.
	 * This estimate is naive:
	 * <ul>
	 *  <li>p_6_17 = minimal non-zero transition probability</li>
	 *  <li>N = number of states</li>
	 * </ul>
	 * <br>
	 * Stores results in consts
	 */
	private void estimatePNRewardQuantile(Constants consts, NondetModel model) throws PrismException
	{
		// we are interested in finding the bound R such that
		// forall s \in S: Pmin[ F{rewards<=R} goal | fail ] > 0
		//
		// TODO: Solve using quantile

		mc.getLog().println("Computing estimates (6.17) via reward bounded quantile...");
		
		Expression goal_or_fail = Expression.Or(new ExpressionLabel(goalLabel), new ExpressionLabel(failLabel));
		ExpressionTemporal ePath = new ExpressionTemporal(ExpressionTemporal.P_F, null, goal_or_fail);
		TemporalOperatorBound bound = new TemporalOperatorBound();
		bound.setBoundType(TemporalOperatorBound.BoundType.REWARD_BOUND);
		bound.setRewardStructureIndex(null);  // no reward structure index = default first reward
		ePath.getBounds().addBound(bound);
		ExpressionProb e = new ExpressionProb(ePath, MinMax.min(), "=", null);
		e.typeCheck();

		NondetModelChecker mc = (NondetModelChecker) this.mc.createModelChecker(model);
		for (int R=1; true; R++) {
			bound.setUpperBound(new ExpressionLiteral(TypeInt.getInstance(), R), false);  // <= R

			mc.getLog().println("Checking " + e + "...");
			JDDNode values = mc.checkExpressionDD(e, model.getReach().copy());
			if (extraVerbose) StateValuesMTBDD.print(mc.getLog(), values.copy(), model, "Pmin for R = " + R);
			JDDNode zeroValues = JDD.Equals(values.copy(), 0);
			if (!JDD.AreIntersecting(zeroValues, model.getReach())) {
				// done!
				double p_min_target = JDD.FindMinPositive(values);

				computeLambda(consts, 2, p_min_target, R);
				consts.details_6_17 = "[via quantile] R = " + R + ", p = " + p_min_target + ", 1-p = " + (1.0-p_min_target); 

				JDD.Deref(zeroValues, values);
				break;
			}

			JDD.Deref(zeroValues, values);
		}
	}

	private void computeLambda(Constants consts, int k, double p_min, int window)
	{
		if (k<=1) {
			throw new IllegalArgumentException("k has to be at least 2, is " + k);
		}

		consts.R0 = (k-1)*window;
		consts.lambda = Math.pow(1.0 - p_min, 1.0/(k*window));

		double oneMinusLambda = 1 - consts.lambda;
		consts.C = 1 / (oneMinusLambda * oneMinusLambda);
	}

	private StateValuesMTBDD computeThetaSaturated(final ComputedValues computed) throws PrismException
	{
		final NondetModel cleanedUp = computed.cleanedUp.getTransformedModel();
		
		NondetModelTransformation transformation = new NondetModelTransformation(cleanedUp) {

			@Override
			public int getExtraStateVariableCount()
			{
				return 0;
			}

			@Override
			public int getExtraActionVariableCount()
			{
				return 0;
			}

			@Override
			public JDDNode getTransformedTrans()
			{
				JDDNode trans = cleanedUp.getTrans().copy();
				trans = JDD.Times(trans, computed.actMax.copy());

				if (extraVerbose) JDD.PrintMinterms(mc.getLog(), cleanedUp.getTrans().copy(), "trans");
				if (extraVerbose) JDD.PrintMinterms(mc.getLog(), computed.actMax.copy(), "actMax");
				if (extraVerbose) JDD.PrintMinterms(mc.getLog(), trans.copy(), "trans (2)");

				// add self-loop from fail state back in
				trans = JDD.ITE(cleanedUp.getLabelDD(failLabel).copy(), cleanedUp.getTrans().copy(), trans);

				if (extraVerbose) JDD.PrintMinterms(mc.getLog(), trans.copy(), "trans (3)");
				return trans;
			}

			@Override
			public JDDNode getTransformedStart()
			{
				return cleanedUp.getStart().copy();
			}

			public JDDNode getTransformedStateReward(JDDNode stateRew) throws PrismException {
				// rew'(s) = rew(s) * pmax(s)
				JDDNode newRew = JDD.Times(stateRew.copy(), computed.pmax.copy());
				return newRew;
			}

			
			@Override
			public JDDNode getTransformedTransReward(JDDNode transRew)
			{
				// rew'(s,a) = rew(s,a) * pmax(s)
				JDDNode newRew = JDD.Times(transRew.copy(), computed.pmax.copy());
				return newRew;
			}

			@Override
			public JDDNode getReachableStates()
			{
				// we need information for *all* states,
				// so we set everything to reachable
				return cleanedUp.getReach().copy();
			}
		};

		mc.getLog().println("\nTransforming MDP to Mmax...");
		NondetModel transformed = computed.cleanedUp.getTransformedModel().getTransformed(transformation);

		if (verbose) {
			try {
				mc.getLog().println("MDP Mmax: " + transformed.infoString());
				transformed.exportToFile(Prism.EXPORT_DOT, true, new File("Mmax.dot"));
				transformed.exportStateRewardsToFile(Prism.EXPORT_PLAIN, new File("Mmax.srew"));
				transformed.exportTransRewardsToFile(Prism.EXPORT_PLAIN, true, new File("Mmax.trew"));
			} catch (FileNotFoundException e) {}
		}

		mc.getLog().println("Computing theta values (maximal expected reward in Mmax)");
		NondetModelChecker mcTransformed = (NondetModelChecker) mc.createModelChecker(transformed);
		ExpressionReward e = new ExpressionReward(new ExpressionTemporal(ExpressionTemporal.R_C, null, null),
		                                          MinMax.max(), "=", null);
		// need information for *all* states, statesOfInterest = reach
		StateValuesMTBDD theta_ = mcTransformed.checkExpression(e, transformed.getReach().copy()).convertToStateValuesMTBDD();
		StateValuesMTBDD theta = new StateValuesMTBDD(theta_.getJDDNode().copy(), cleanedUp);
		theta_.clear();

		transformed.clear();

		return theta;
	}

	private interface BufferPrintable {
		void print(StringBuffer buffer, String indent);

		default void print(StringBuffer buffer)
		{
			print(buffer, "");
		}

		default String bufferToString()
		{
			StringBuffer buffer = new StringBuffer();
			print(buffer);
			return buffer.toString();
		}
	}


	public static class OptimalChoice implements BufferPrintable {
		private int s;
		private int r;
		private int choice;

		private double p;
		private double theta;

		public OptimalChoice(int s, int r, int choice, double p, double theta)
		{
			this.s = s;
			this.r = r;
			this.choice = choice;
			this.p = p;
			this.theta = theta;
		}

		public int getS()
		{
			return s;
		}

		public int getR()
		{
			return r;
		}

		public int getChoice()
		{
			return choice;
		}

		public String getChoiceString()
		{
			switch (choice) {
			case TAU_CHOICE:
				return "τ";
			case TRAP_CHOICE:
				return "↩";
			default:
				return Integer.toString(choice);
			}
		}

		public double getP()
		{
			return p;
		}

		public double getTheta()
		{
			return theta;
		}

		// Get the conditional expection if this were the initial state
		public double getCE()
		{
			return getTheta() / getP();
		}
		
		public double getDelta(double threshold, int R)
		{
			return getTheta() - ((threshold - R) * getP());
		}

		public String toString()
		{
			return toString(false);
		}

		public String toString(boolean verbose)
		{
			return "act=" + getChoiceString() + ", p=" + getP() + ", θ=" + getTheta() + (verbose ? (", s=" + s + ", r=" + r ): "");
		}

		@Override
		public int hashCode()
		{
			final int prime = 31;
			int result = 1;
			result = prime * result + choice;
			long temp;
			temp = Double.doubleToLongBits(p);
			result = prime * result + (int) (temp ^ (temp >>> 32));
			result = prime * result + r;
			result = prime * result + s;
			temp = Double.doubleToLongBits(theta);
			result = prime * result + (int) (temp ^ (temp >>> 32));
			return result;
		}

		@Override
		public boolean equals(Object obj)
		{
			if (this == obj)
				return true;
			if (obj == null)
				return false;
			if (getClass() != obj.getClass())
				return false;
			OptimalChoice other = (OptimalChoice) obj;
			if (choice != other.choice)
				return false;
			if (Double.doubleToLongBits(p) != Double.doubleToLongBits(other.p))
				return false;
			if (r != other.r)
				return false;
			if (s != other.s)
				return false;
			if (Double.doubleToLongBits(theta) != Double.doubleToLongBits(other.theta))
				return false;
			return true;
		}

		@Override
		public void print(StringBuffer buffer, String indent)
		{
			buffer.append(indent);
			buffer.append(toString());
			buffer.append("\n");
		}
		
		public static OptimalChoice forGoal(int s, int r)
		{
			return new OptimalChoice(s, r, TRAP_CHOICE, 1.0, 0.0);
		}

		public static OptimalChoice forFail(int s, int r)
		{
			return new OptimalChoice(s, r, TRAP_CHOICE, 0.0, 0.0);
		}
	}

	@SuppressWarnings("serial")
	private class OptimalChoicesForStates<T extends BufferPrintable> extends ArrayList<T> implements BufferPrintable
	{
		public OptimalChoicesForStates(int numStates)
		{
			super(numStates);
		}

		public T getFor(int s)
		{
			return get(s);
		}

		public void print(StringBuffer buffer, String indent)
		{
			for (int s=0; s < size(); s++) {
				buffer.append("State " + s + ":\n");
				if (getFor(s) == null) 
					buffer.append(indent + "  ?\n");
				else {
					getFor(s).print(buffer, indent + "  ");
				}
				//buffer.append("\n");
			}
		}

		public String toString()
		{
			StringBuffer buffer = new StringBuffer();
			print(buffer);
			return buffer.toString();
		}
	}

	private interface OptimalChoicesComputed<T extends BufferPrintable> extends BufferPrintable {
		public T getFor(int r, int s);
		public int getRequiredMemory();
		
		public Iterable<Integer> getLevelsDescending();
		public Iterable<Integer> getStatesForLevel(int r);
	}

	@SuppressWarnings("serial")
	private class OptimalChoices<T extends BufferPrintable>
		extends ArrayList<OptimalChoicesForStates<T>>
		implements BufferPrintable, OptimalChoicesComputed<T>
	{
		private int numStates;

		public OptimalChoices(int numStates, int R)
		{
			super(R+1);
			this.numStates = numStates;
			for (int r = 0; r < R+1; r++) {
				this.add(new OptimalChoicesForStates<>(numStates));
			}
		}

		@Override
		public Iterable<Integer> getLevelsDescending()
		{
			return new Iterable<Integer>() {

				@Override
				public Iterator<Integer> iterator()
				{
					return new Iterator<Integer>() {
						int cur = size() - 1;

						@Override
						public boolean hasNext()
						{
							return cur >= 0;
						}

						@Override
						public Integer next()
						{
							int result = cur;
							cur--;
							return result;
						}
					};
				}
			};
		}

		@Override
		public Iterable<Integer> getStatesForLevel(int r)
		{
			return new common.iterable.Interval(0, numStates);
		}

		@Override
		public int getRequiredMemory() {
			return size();
		}

		public OptimalChoicesForStates<T> getFor(int r)
		{
			return get(r);
		}

		public T getFor(int r, int s)
		{
			return getFor(r).getFor(s);
		}

		public void setFor(int r, int s, T t)
		{
			getFor(r).set(s, t);
		}

		
		public void print(StringBuffer buffer, String indent)
		{
			print(buffer, indent, 0);
		}

		public void print(StringBuffer buffer, String indent, int fromLevel)
		{
			for (int r=this.size()-1; r>=fromLevel; r--) {
				buffer.append(indent + "----- Level " + r + ": --------------------\n");
				getFor(r).print(buffer, indent);
				buffer.append("\n");
			}
		}

		public String toString()
		{
			return bufferToString();
		}
		
		public String toString(int fromLevel)
		{
			StringBuffer buffer = new StringBuffer();
			print(buffer, "", fromLevel);
			return buffer.toString();
		}
	}

	@SuppressWarnings("serial")
	private static class OptimalChoiceMap implements BufferPrintable, OptimalChoicesComputed<OptimalChoice>
	{
		TreeMap<Integer, TreeMap<Integer, OptimalChoice>> map;
		int numStates;
		
		public OptimalChoiceMap(int numStates)
		{
			map = new TreeMap<Integer, TreeMap<Integer, OptimalChoice>>();
			this.numStates = numStates;
		}

		
		@Override
		public Iterable<Integer> getLevelsDescending()
		{
			return map.descendingKeySet();
		}
		
		public Iterable<Integer> getStatesForLevel(int r)
		{
			return getFor(r).keySet();
		}

		public OptimalChoice getFor(int r, int s)
		{
			TreeMap<Integer, OptimalChoice> choicesForR = getFor(r);
			if (choicesForR != null) {
				return choicesForR.get(s);
			}
			return null;
		}

		public TreeMap<Integer, OptimalChoice> getFor(int r)
		{
			return map.get(r);
		}

		public void setFor(int r, int s, OptimalChoice c)
		{
			TreeMap<Integer, OptimalChoice> choicesForR = getFor(r);
			if (choicesForR == null) {
				choicesForR = new TreeMap<Integer, OptimalChoice>();
				map.put(r, choicesForR); 
			}
			if (getFor(r).put(s, c) != null) {
				throw new RuntimeException("Internal error: Replacing existing optimal choice");
			}
		}

		public void print(StringBuffer buffer, String indent)
		{
			for (int r : map.navigableKeySet().descendingSet()) {
				buffer.append(indent + "----- Level " + r + ": --------------------\n");
				for (Entry<Integer, OptimalChoice> e : getFor(r).entrySet()) {
					buffer.append("State " + e.getKey() + ":\n");
					e.getValue().print(buffer, indent + "  ");
				}
				buffer.append("\n");
			}
		}

		public String toString()
		{
			return bufferToString();
		}

		@Override
		public int getRequiredMemory()
		{
			return map.lastKey() + 1;
		}
	}

	private double computeCEMaxSequential(ThresholdAlgorithm info) throws PrismException
	{
		int statThresholdCalls = 0;
		mc.getLog().println("\nComputing CEmax via sequential calls to the threshold algorithm (" + info.getName() +")...\n");
		StopWatch timerCEMax = new StopWatch(mc.getLog());
		timerCEMax.start("CEmax computation");
		long timerThreshold;
		long timeThreshold = 0;

		double epsilon = mc.getSettings().getDouble(PrismSettings.PRISM_TERM_CRIT_PARAM);
		boolean epsilonAbsolute = mc.getSettings().getString(PrismSettings.PRISM_TERM_CRIT).equals("Absolute");

		// initialise with M scheduler CE
		double A = info.getCEforM();

		while (true) {
			statThresholdCalls++;
			mc.getLog().println("[compute for threshold = " + A+ ", iteration = " + statThresholdCalls+"] call CEmax >= " + A + "?");
			timerThreshold = System.currentTimeMillis();
			evaluateThresholdLower(info, A, false);
			timerThreshold = System.currentTimeMillis() - timerThreshold;
			timeThreshold += timerThreshold;

			double Anew = info.getCEforSinit();
			mc.getLog().println("[compute for threshold = " + A+ ", iteration = " + statThresholdCalls +"] result CEsched = " + (Double.isNaN(Anew) ? "NO" : Anew) + ", took " + timerThreshold/1000.0 + "s.");
			if (Double.isNaN(Anew)) {
				throw new PrismException("Obtained NO response, should not happen for sequential CEmax computation. Numerical instability?");
			}
			if (Anew == A || doublesAreClose(Anew, A, epsilon, epsilonAbsolute)) {
				mc.getLog().println("Found CEsched with CEsched = CEmax" +
				                     ( Anew == A ? "" : " (up to epsilon)")+ ", finished.");
				break;
			}
			if (A > Anew) {
				throw new PrismException("Old CEsched > new CEsched, should not happen: " + A + ", " + Anew + ", probably numerical instability");
			}
			A = Anew;
		}

		mc.getLog().println();
		timerCEMax.stop();
		mc.getLog().println("Called threshold subroutine " + statThresholdCalls + " times, combined time for threshold computations: " + timeThreshold/1000.0 +"s.");

		return A;
	}

	private double computeCEMax(ThresholdAlgorithm info, double upper) throws PrismException
	{
		int statThresholdCalls = 0;
		mc.getLog().println("\nComputing CEmax via scheduler improvement (" + info.getName() +")...\n");
		StopWatch timerCEMax = new StopWatch(mc.getLog());
		timerCEMax.start("CEmax computation");
		long timerThreshold;
		long timeThreshold = 0;

		double epsilon = mc.getSettings().getDouble(PrismSettings.PRISM_TERM_CRIT_PARAM);
		boolean epsilonAbsolute = mc.getSettings().getString(PrismSettings.PRISM_TERM_CRIT).equals("Absolute");

		double A = info.getCEforM();
		double B = upper + 0.1;  // to account for numerical instability

		double CEsched;
		String prefix = "[compute against CE for M, A=" + A + ", B="+B+"] ";

		mc.getLog().println(prefix + "Initialize search for CEmax, interval: ["+A+","+B+"[");
		mc.getLog().println(prefix + "Run CEmax >= " + A);
		timerThreshold = System.currentTimeMillis();
		evaluateThresholdLower(info, A, false);
		timeThreshold += System.currentTimeMillis() - timerThreshold;
		statThresholdCalls++;
		CEsched = info.getCEforSinit();
		mc.getLog().println(prefix + "Result for CEmax >= " + A + ": " + CEsched);

		if (CEsched == A || doublesAreClose(CEsched, A, epsilon, epsilonAbsolute)) {
			mc.getLog().println(prefix + "Found CEsched with CEsched = CEmax" +
			                    ( CEsched == A ? "" : " (up to epsilon)")+ ", finished.");
			timerCEMax.stop();
			mc.getLog().println("Called threshold subroutine " + statThresholdCalls + " times, combined time for threshold computations: " + timeThreshold/1000.0 +"s.");
			return CEsched;
		}
		A = CEsched;

		int r = info.getHighestNonOptimalLevel();
		prefix = "[compute, r="+r+", A=" + A + ", B="+B+"] ";

		outer: while (r >= 0) {
			TreeSet<Double> I = info.generateI(r);
			NavigableSet<Double> I_B = I.subSet(0.0, true, B, false);  // I_B = { d in I : d < B }
			NavigableSet<Double> I_up = I.subSet(CEsched, true, Double.POSITIVE_INFINITY, false);

			if (I_B.isEmpty()) {
				info.setOptimalFromLevel(r);
				mc.getLog().println(prefix + "I_B is empty => optimal from level r = " + r + " above");
				r--;
				prefix = "[compute, r="+r+", A=" + A + ", B="+B+"] ";
				continue outer;
			}

			prefix = "[compute, r="+r+", A=" + A + ", B="+B+"] ";

			TreeSet<Double> K = new TreeSet<Double>(I_up);
			K.add(CEsched);

			inner: while (true) {
				NavigableSet<Double> J = K.subSet(A, true, B, false);  // J = K \cap I
				if (J.isEmpty()) {
					// TODO: Handle better?!
					throw new PrismException("Encountered case where K ∩ I = empty even though I_B is non-empty. This might be due to numerical imprecision, aborting...");
				}
				// mc.getLog().println("K: " + K);
				// mc.getLog().println("J: " + J);
				// if (J.isEmpty()) {}
				double threshold = J.last();  // next threshold to check = max K \cap I

				mc.getLog().println(prefix + "Run CEmax >= " + threshold);
				timerThreshold = System.currentTimeMillis();
				evaluateThresholdLower(info, threshold, false);
				timeThreshold += System.currentTimeMillis() - timerThreshold;
				statThresholdCalls++;

				CEsched = info.getCEforSinit();
				// CEsched == NaN   iff  info.getPforSinit() = 0
				if (Double.isNaN(CEsched) && info.getPforSinit() != 0) {
					throw new PrismException("Got NaN for CEsched, but probability is != 0, probably numerical instability?");
				}
				mc.getLog().println(prefix + "Result for CEmax >= " + threshold + ": " + (Double.isNaN(CEsched) ? "NO" : CEsched));

				if (CEsched == threshold || doublesAreClose(CEsched, threshold, epsilon, epsilonAbsolute)) {
					mc.getLog().println(prefix + "Found CEsched with CEsched = CEmax" +
					                    ( CEsched == threshold ? "" : " (up to epsilon)")+ ", finished.");
					timerCEMax.stop();
					mc.getLog().println("Called threshold subroutine " + statThresholdCalls + " times, combined time for threshold computations: " + timeThreshold/1000.0 +"s.");
					return CEsched;
				}

				if (Double.isNaN(CEsched)) {
					// NO response
					B = threshold;
					mc.getLog().println(prefix + "(case NO) Update upper bound to " + B);
				} else if (CEsched < threshold) {
					B = threshold;
					if (CEsched <= A) {
						B = threshold; 
						mc.getLog().println(prefix + "(case CEsched'<=A) Update upper bound to " + B);
					} else {
						// CEsched > A
						A = CEsched;
						B = threshold;
						K.add(CEsched);
						mc.getLog().println(prefix + "(case CEsched'>A) Update lower bound to " + A + ", upper bound to " + B);
					}
				} else {
					// CEsched > threshold
					A = CEsched;
					mc.getLog().println(prefix + "(case CEsched'>threshold) Update lower bound to " + A);
					continue outer;
				}
				prefix = "[compute, r="+r+", A=" + A + ", B="+B+"] ";
				continue inner;
			}
		}

		mc.getLog().println();
		timerCEMax.stop();
		mc.getLog().println("Called threshold subroutine " + statThresholdCalls + " times, combined time for threshold computations: " + timeThreshold/1000.0 +"s.");
		return info.getCEforSinit();
	}

	private abstract class ThresholdAlgorithm
	{
		final MDPSparse mdp;
		final MDPRewardsSimple rew;
		final int sInit;

		final int numStates;
		protected List<BitSet> sccs;

		final explicit.StateValues P_M;  // y for deterministic scheduler M
		final explicit.StateValues E_M;  // theta for deterministic scheduler M
	
		protected BitSet goal;
		protected BitSet fail;

		protected Integer optimalFromLevel;

		ThresholdAlgorithm(MDPSparse mdp, MDPRewardsSimple rew, explicit.StateValues P_M, explicit.StateValues E_M) throws PrismException
		{
			this.mdp = mdp;
			this.rew = rew;

			this.P_M = P_M;
			this.E_M = E_M;

			numStates = mdp.getNumStates();

			goal = mdp.getLabelStates(goalLabel);
			fail = mdp.getLabelStates(failLabel);

			sInit = mdp.getInitialStates().iterator().next();
		}

		public abstract String getName();

		public void setOptimalFromLevel(int r)
		{
			optimalFromLevel = r;
		}

		public abstract OptimalChoicesComputed<OptimalChoice> getOpt();

		/**
		 * Return the highest non optimal level, -1 if all levels are optimal.
		 */
		public abstract int getHighestNonOptimalLevel();

		public double getCEforM()
		{
			return (double)E_M.getValue(sInit) / (double)P_M.getValue(sInit);
		}

		public abstract double getCEforSinit();
		public abstract double getPforSinit();

		public abstract TreeSet<Double> generateI(int r) throws PrismException;

		public void generateI(int s, int r, Integer R, TreeSet<Double> result) throws PrismException
		{
			if (mdp.getNumChoices(s) <= 1) {
				return;
			}
			OptimalChoice selected = getOpt().getFor(r, s);
			if (verbose) mc.getLog().println("Selected choice = " + selected);
			for (int choice = 0; choice < mdp.getNumChoices(s); choice++) {
				if (selected.getChoice() == choice)
					continue;

				OptimalChoice alpha = computeOptimalChoice(s, r, choice, getOpt(), mdp, rew, R);
				double split = r + (selected.getTheta() - alpha.getTheta()) / (selected.getP() - alpha.getP());
				if (selected.getP() <= alpha.getP()) {
					if (verbose) mc.getLog().println("Ignoring split = " + split + " for alpha " + alpha);
					continue;
				}

				if (verbose) mc.getLog().println("Adding split = " + split + " for alpha " + alpha);

				result.add(split);
			}
		}

	}

	private class ThresholdAlgorithmAcyclic extends ThresholdAlgorithm {
		OptimalChoiceMap opt;
		
		ThresholdAlgorithmAcyclic(MDPSparse mdp, MDPRewardsSimple rew, explicit.StateValues P_M, explicit.StateValues E_M) throws PrismException
		{
			super(mdp, rew, P_M, E_M);
			opt = new OptimalChoiceMap(mdp.getNumStates());
		}

		@Override
		public double getCEforSinit()
		{
			return opt.getFor(0, sInit).getCE();
		}

		@Override
		public double getPforSinit()
		{
			return opt.getFor(0, sInit).getP();
		}

		@Override
		public int getHighestNonOptimalLevel()
		{
			if (optimalFromLevel == null) {
				return opt.map.lastKey();
			}
			Integer level = opt.map.floorKey(optimalFromLevel - 1);
			if (level == null)
				return -1;
			return level;
		}

		@Override
		public OptimalChoicesComputed<OptimalChoice> getOpt()
		{
			return opt;
		}

		@Override
		public TreeSet<Double> generateI(int r) throws PrismException
		{
			BitSet goal = mdp.getLabelStates(goalLabel);
			BitSet fail = mdp.getLabelStates(failLabel);

			if (verbose) {
				mc.getLog().println("Computing I for level r = " +r);
			}

			TreeMap<Integer, OptimalChoice> stored = opt.getFor(r);

			TreeSet<Double> result = new TreeSet<>();
			for (int s : stored.keySet()) {
				if (goal.get(s) || fail.get(s)) {
					continue;
				}
				generateI(s, r, null, result);
			}

			return result;
		}

		public void clearNonOptimal()
		{
			if (optimalFromLevel == null) {
				opt.map.clear();
			} else {
				while (!opt.map.isEmpty()) {
					int i = opt.map.firstKey();
					if (i < optimalFromLevel) {
						opt.map.remove(i);
					} else {
						break;
					}
				}
			}
		}

		@Override
		public String getName()
		{
			return "acyclic case";
		}
	}

	private class ThresholdAlgorithmGeneralCase extends ThresholdAlgorithm {
		final int R;

		OptimalChoices<OptimalChoice> opt;

		public ThresholdAlgorithmGeneralCase(MDPSparse mdp, MDPRewardsSimple rew, int R, explicit.StateValues P_M, explicit.StateValues E_M) throws PrismException
		{
			super(mdp, rew, P_M, E_M);

			this.R = R;

			optimalFromLevel = R;

			init();
		}

		public void init() throws PrismException
		{
			mc.getLog().println("Allocating data structures for threshold algorithm, " + numStates + " states and " + (R+1) + " reward levels");
			StopWatch timer = new StopWatch(mc.getLog());
			timer.start("allocating data structures and obtaining topological order");

			opt = new OptimalChoices<OptimalChoice>(numStates, R);

			for (int r = 0; r <= R; r++) {
				for (int s = 0; s < numStates; s++) {
					opt.getFor(r).add(null);
				}
			}

			// initialize normal states on all layers
			for (int s = 0; s < numStates; s++) {
				for (int r = 0; r <= R; r++) {
					OptimalChoice ch = new OptimalChoice(s, r, TAU_CHOICE, (double)P_M.getValue(s), (double)E_M.getValue(s));
					opt.setFor(r, s, ch);
				}
			}
		
			// initialize goal / fail
			for (int s : new IterableBitSet(goal)) {
				for (int r = 0; r <= R; r++) {
					OptimalChoice ch = OptimalChoice.forGoal(s, r);
					opt.setFor(r, s, ch);
				}
			}
			for (int s : new IterableBitSet(fail)) {
				for (int r = 0; r <= R; r++) {
					OptimalChoice ch = OptimalChoice.forFail(s, r);
					opt.setFor(r, s, ch);
				}
			}

			PairPredicateInt dropPositiveRewTrans = new PairPredicateInt() {
				@Override
				public boolean test(int state, int choice)
				{
					return rew.getTransitionReward(state, choice) > 0;
				}
			};

			MDPDroppedChoices zeroRewardMDP = new MDPDroppedChoices(mdp, dropPositiveRewTrans);
			if (verbose) {
				zeroRewardMDP.exportToDotFile("explicit-zero-reward.dot");
			}

			mc.getLog().println("Compute SCCs in MDP to obtain topological order of states...");
			SCCConsumerStore sccStore = new SCCConsumerStore(mc, zeroRewardMDP);
			explicit.SCCComputer sccComputer = explicit.SCCComputer.createSCCComputer(prism, zeroRewardMDP, sccStore);
			sccComputer.computeSCCs(false);  // don't filter trivial SCCs

			sccs = sccStore.getSCCs();

			timer.stop();
		}

		@Override
		public int getHighestNonOptimalLevel()
		{
			return optimalFromLevel - 1;
		}

		public OptimalChoicesComputed<OptimalChoice> getOpt()
		{
			return opt;
		}

		@Override
		public double getCEforSinit()
		{
			return opt.getFor(0, sInit).getCE();
		}

		@Override
		public double getPforSinit()
		{
			return opt.getFor(0, sInit).getP();
		}

		public TreeSet<Double> generateI(int r) throws PrismException
		{
			BitSet goal = mdp.getLabelStates(goalLabel);
			BitSet fail = mdp.getLabelStates(failLabel);

			if (verbose) {
				mc.getLog().println("Computing I for level r = " +r);
			}

			TreeSet<Double> result = new TreeSet<>();
			for (int s = 0; s < mdp.getNumStates(); s++) {
				if (goal.get(s) || fail.get(s)) {
					continue;
				}
				generateI(s, r, R, result);
			}

			return result;
		}

		@Override
		public String getName()
		{
			return "general case, saturation point = " + R;
		}

	};
	
	private OptimalChoicesComputed<OptimalChoice> evaluateThresholdLower(ThresholdAlgorithm info, final double threshold, final double lowerBound, final double upperBound, boolean strict, final Integer R) throws PrismException
	{
		boolean upperBoundClose = doublesAreClose(threshold, upperBound,
		                                          mc.getSettings().getDouble(PrismSettings.PRISM_TERM_CRIT_PARAM),
		                                          true);
		if (strict) {
			if (threshold >= upperBound && !upperBoundClose) {
				mc.getLog().println("Threshold " + threshold + " >= known upper bound " + upperBound + ", trivially false.");
				return null;
			}
		} else {
			if (threshold > upperBound && !upperBoundClose) {
				mc.getLog().println("Threshold " + threshold + " > known upper bound " + upperBound + ", trivially false.");
				return null;
			}
		}

		int numStates = info.mdp.getNumStates();

		// there is a single initial state
		
		boolean checkSchedMax = false;
		
		// is the deterministic, memoryless scheduler in SchedMax already over the threshold?
		OptimalChoice init = new OptimalChoice(info.sInit, 0, TAU_CHOICE, (double)info.P_M.getValue(info.sInit), (double)info.E_M.getValue(info.sInit));
		if (checkSchedMax && (strict ? init.getCE() > threshold : init.getCE() >= threshold)) {
			// yay, construct witness

			OptimalChoices<OptimalChoice> opt = new OptimalChoices<OptimalChoice>(numStates, 0);

			for (int s = 0; s < numStates; s++) {
				opt.getFor(0).add(null);
			}
		
			BitSet goal = info.mdp.getLabelStates(goalLabel);
			BitSet fail = info.mdp.getLabelStates(failLabel);

			// set normal states to TAU_CHOICE
			for (int s = 0; s < numStates; s++) {
				OptimalChoice ch = new OptimalChoice(s, 0, TAU_CHOICE, (double)info.P_M.getValue(s), (double)info.E_M.getValue(s));
				opt.setFor(0, s, ch);
			}
		
			// set goal / fail
			for (int s : new IterableBitSet(goal)) {
				OptimalChoice ch = OptimalChoice.forGoal(s, 0);
				opt.setFor(0, s, ch);
			}
			for (int s : new IterableBitSet(fail)) {
				OptimalChoice ch = OptimalChoice.forFail(s, 0);
				opt.setFor(0, s, ch);
			}

			return opt;
		}

		if (verbose) {
			mc.getLog().println("\n\n==== after initialization =====\n");
			mc.getLog().print(info.getOpt());
		}

		evaluateThresholdLower(info, threshold, strict);

		mc.getLog().println("Potential witness/counter-example has CE of " + info.getCEforSinit());

		if (Double.isNaN(info.getCEforSinit())) {
			// NO response due to p=0.0
			return null;
		}
		if (strict ? info.getCEforSinit() > threshold : info.getCEforSinit() >= threshold) {
			// yay, found a witness
			return info.getOpt();
		} else {
			return null;
		}
	}
	
	private void evaluateThresholdLower(ThresholdAlgorithm info, double threshold, boolean strict) throws PrismException
	{
		if (info instanceof ThresholdAlgorithmGeneralCase) {
			evaluateThresholdLower((ThresholdAlgorithmGeneralCase)info, threshold, strict);
		} else {
			evaluateThresholdLower((ThresholdAlgorithmAcyclic)info, threshold, strict);
		}
	}

	private void evaluateThresholdLower(ThresholdAlgorithmGeneralCase info, double threshold, boolean strict) throws PrismException
	{
		for (int r = info.optimalFromLevel-1; r >= 0; r--) {
			// go level by level

			if (verbose) {
				mc.getLog().println("\n\n==== computing (r = " + r +") =====\n");
			}

			for (BitSet scc : info.sccs) {
				if (scc.intersects(info.goal) || scc.intersects(info.fail)) {
					continue;
				}
				if (scc.cardinality() > 1) {
					// zero reward fragment...
					optimizeZeroRewardFragment(info.mdp, info.rew, scc, r, strict, info.opt, info.R, threshold);
				} else {
					// single state
					int s = scc.nextSetBit(0);
					OptimalChoice feasible;

					feasible = computeFeasible(info.mdp, info.rew, s, r, strict, info.opt, info.R, threshold);
					info.opt.setFor(r, s, feasible);
				}
			}

			if (verbose) {
				mc.getLog().println();
				mc.getLog().println(info.opt.toString(r));
			}
		}
	}

	private static final class SRStackContent
	{
		int s;
		int r;
		boolean firstPhase;

		public SRStackContent(int s, int r) {
			this.s = s;
			this.r = r;
			this.firstPhase = true;
		}
		
		public void nextPhase()
		{
			firstPhase = false;
		}
	}

	private void evaluateThresholdLower(ThresholdAlgorithmAcyclic info, double threshold, boolean strict) throws PrismException
	{
		info.clearNonOptimal();

		Stack<SRStackContent> stack = new Stack<>();

		stack.push(new SRStackContent(info.sInit, 0));

		while (!stack.isEmpty()) {
			SRStackContent cur = stack.pop();

			int s = cur.s;
			int r = cur.r;
			boolean firstPhase = cur.firstPhase;

			if (info.opt.getFor(r, s) != null) {
				// already computed
				continue;
			}

			if (info.goal.get(s)) {
				OptimalChoice ch = OptimalChoice.forGoal(s, r);
				info.opt.setFor(r, s, ch);
				continue;
			}

			if (info.fail.get(s)) {
				OptimalChoice ch = OptimalChoice.forFail(s, r);
				info.opt.setFor(r, s, ch);
				continue;
			}

			if (cur.firstPhase) {
				cur.nextPhase();
				stack.push(cur);
			}

			if (firstPhase) {
				for (int ch = 0, numChoices = info.mdp.getNumChoices(); ch < numChoices; ch++) {
					for (Iterator<Integer> it = info.mdp.getSuccessorsIterator(cur.s, ch); it.hasNext(); ) {
						int t = it.next();
						int rew = SafeCast.toInt(info.rew.getStateReward(cur.s)) + SafeCast.toInt(info.rew.getTransitionReward(cur.s, ch));
						rew += r;

						if (info.opt.getFor(rew, t) == null) {
							stack.push(new SRStackContent(t, rew));
						}
					}
				}
			} else {
				// all successors have been computed
				OptimalChoice choice = computeFeasible(info.mdp, info.rew, s, r, strict, info.opt, null, threshold);
				info.opt.setFor(r, s, choice);
			}
		}
	}

	/** Do absolute / relative comparison, but use relative only for values &lt;=1 */ 
	private boolean doublesAreClose(double a, double b, double epsilon, boolean absolute)
	{
		if (a > 1) {
			// divisor > 1, relative comparison would be *less* precise than absolute,
			// so we switch to absolute
			absolute = true;
		}
		return PrismUtils.doublesAreClose(a, b, epsilon, absolute);
	}

	private void optimizeZeroRewardFragment(MDPSparse mdp, MDPRewardsSimple rew, BitSet scc, int r, boolean strict, OptimalChoices<OptimalChoice> opt, int R, double threshold) throws PrismException
	{
		throw new PrismNotSupportedException("Currently, can not handle zero-reward cycles: " + scc);
	}

	/**
	 * Determines best feasible action from list of choices.
	 * <br>
	 * Note: The choices list is modified during exection!
	 */
	private OptimalChoice computeBestFeasibleF1F2(int r, ArrayList<OptimalChoice> choices, boolean strict, double threshold) throws PrismException
	{
		Comparator<OptimalChoice> descending = (OptimalChoice a, OptimalChoice b) -> {
			if (a.getP() == b.getP()) {
				return Double.compare(a.getTheta(), b.getTheta());
			}
			if (a.getP() > b.getP()) {
				return -1;
			}
			return 1;
		};

		choices.sort(descending);

		// (F2): remove all actions with y_1 = y_2 and same/smaller theta but one
		OptimalChoice last = null;
		for (Iterator<OptimalChoice> it = choices.iterator(); it.hasNext(); ) {
			OptimalChoice cur = it.next();
			if (last != null && last.getP() == cur.getP()) {
				assert(last.getTheta() >= cur.getTheta()); 
				it.remove();
			}
		}

		OptimalChoice currentBest = choices.get(0);
		for (int i=1; i < choices.size(); i++) {
			OptimalChoice beta = choices.get(i);
			assert (currentBest.getP() > beta.getP());

			double p_diff = currentBest.getP() - beta.getP();
			double theta_diff = currentBest.getTheta() - beta.getTheta();
			double quotient = theta_diff / p_diff;
			double r_plus_quotient = r + quotient;
			
			if (verbose) {
				mc.getLog().println("choice alpha = " + currentBest.getChoiceString() + ", choice beta = " + beta.getChoiceString());
				mc.getLog().println("theta alpha  = " + currentBest.getTheta() + "\t theta beta = " + beta.getTheta());
				mc.getLog().println("p alpha      = " + currentBest.getP()     + "\t p beta     = " + beta.getP());
				mc.getLog().println("diff theta   = " + theta_diff       + "\t diff p     = " + p_diff);
				mc.getLog().println("split = " + r_plus_quotient + "      =  " + r + " + " + theta_diff / p_diff   + "     >? " + threshold);
			}

			if (!(strict ? (r_plus_quotient > threshold) : (r_plus_quotient >= threshold))) {
				currentBest = beta;

				if (verbose) mc.getLog().println("-> Update currentBest = " + currentBest.getChoice());
			} else {
				if (verbose) mc.getLog().println("-> Keep currentBest = " + currentBest.getChoice());
			}
		}

		return currentBest; 
	}

	/**
	 * Determines best feasible action from list of choices (delta).
	 */
	private OptimalChoice computeBestFeasibleDelta(int r, ArrayList<OptimalChoice> choices, boolean strict, double threshold) throws PrismException
	{
		OptimalChoice currentBest = choices.get(0);
		for (int i=1; i < choices.size(); i++) {
			OptimalChoice beta = choices.get(i);

			double delta_alpha = currentBest.getDelta(threshold, r);
			double delta_beta = beta.getDelta(threshold, r);

			if (verbose) {
				mc.getLog().println("choice alpha = " + currentBest.getChoiceString() + ",\t choice beta = " + beta.getChoiceString());
				mc.getLog().println("theta alpha  = " + currentBest.getTheta() + "\t theta beta = " + beta.getTheta());
				mc.getLog().println("p alpha      = " + currentBest.getP()     + "\t p beta     = " + beta.getP());
				mc.getLog().println("delta alpha  = " + delta_alpha            + "\t delta beta = " + delta_beta);
			}

			if (delta_beta > delta_alpha) {
				currentBest = beta;
				if (verbose) mc.getLog().println("-> Update currentBest = " + currentBest.getChoice() + " (better delta)");
			} else if (delta_beta == delta_alpha) {
				if (currentBest.getP() < beta.getP()) {
					currentBest = beta;
					if (verbose) mc.getLog().println("-> Update currentBest = " + currentBest.getChoice() + " (same delta, better y)");
				} else {
					if (verbose) mc.getLog().println("-> Keep currentBest = " + currentBest.getChoice() + " (same delta, better y)");
				}
			} else {
				if (verbose) mc.getLog().println("-> Keep currentBest = " + currentBest.getChoice() + " (better delta)");
			}
		}

		return currentBest; 
	}

	private OptimalChoice computeFeasible(MDPSparse mdp, MDPRewardsSimple rew, int s, int r, boolean strict, OptimalChoicesComputed<OptimalChoice> opt, Integer R, double threshold) throws PrismException
	{
		ArrayList<OptimalChoice> choices = new ArrayList<OptimalChoice>(mdp.getNumChoices(s));
		for (int choice = 0; choice < mdp.getNumChoices(s); choice++) {
			OptimalChoice alpha = computeOptimalChoice(s, r, choice, opt, mdp, rew, R);
			choices.add(alpha);
		}

		// return computeBestFeasibleF1F2(r, choices, strict, threshold);
		return computeBestFeasibleDelta(r, choices, strict, threshold);
	}

	private OptimalChoice computeOptimalChoice(int s, int r, int alpha, OptimalChoicesComputed<OptimalChoice> opt, MDPSparse mdp, MDPRewardsSimple rew, Integer R) throws PrismException
	{
		double p = 0.0;
		double theta = 0.0;

		int transRew = SafeCast.toInt(rew.getTransitionReward(s, alpha));

		if (transRew == 0 && hasSelfloopForAction(mdp, s, alpha)) {
			return computeOptimalChoiceWithSelfloop(s, r, alpha, opt, mdp);
		}

		if (verbose)
			mc.getLog().println("Optimal values for s=" + s + ", r="+ r + " and act=" + alpha + " and transition reward " + transRew);

		int rPrime = r + transRew;
		if (R != null && rPrime > R) {  // R = null => no saturation point specified
			rPrime = R;
		}

		for (Iterator<Entry<Integer, Double>> it = mdp.getTransitionsIterator(s, alpha); it.hasNext(); ) {
			Entry<Integer, Double> e = it.next();
			int t = e.getKey();
			Double prob = e.getValue();

			OptimalChoice ch = opt.getFor(rPrime, t);

			if (ch == null) {
				throw new PrismException("Value not computed yet...");
			}

			p += prob * ch.getP();
			theta += prob * ch.getTheta();
			if (extraVerbose) {
				mc.getLog().println(ch + "  // t = " + t + ", R' = " + rPrime);
				mc.getLog().println("p += " + prob + " * " + ch.getP() + " = " + (prob * ch.getP()) + " // t = " + t);
				mc.getLog().println("theta += " + prob + " * " + ch.getTheta() + " = " + (prob * ch.getTheta()) + " // t = " + t);
			}
		}
		if (verbose) mc.getLog().print("p = " + p);
		if (extraVerbose) {
			mc.getLog().print(", theta_ = " + theta);
		}

		theta += p * transRew;
		if (verbose) mc.getLog().println(", theta = " + theta);

		return new OptimalChoice(s, r, alpha, p, theta);
	}

	private OptimalChoice computeOptimalChoiceWithSelfloop(int s, int r, int alpha, OptimalChoicesComputed<OptimalChoice> opt, MDPSparse mdp) throws PrismException
	{
		double p = 0.0;
		double theta = 0.0;

		if (verbose)
		mc.getLog().println("Optimal values for s=" + s + ", r="+ r + " and act=" + alpha + " and transition reward 0 (with selfloop)");

		double selfProb = 0.0;
		
		for (Iterator<Entry<Integer, Double>> it = mdp.getTransitionsIterator(s, alpha); it.hasNext(); ) {
			Entry<Integer, Double> e = it.next();

			int t = e.getKey();
			Double prob = e.getValue();
			
			if (t == s) {
				selfProb += prob;
			}
		}

		for (Iterator<Entry<Integer, Double>> it = mdp.getTransitionsIterator(s, alpha); it.hasNext(); ) {
			Entry<Integer, Double> e = it.next();
			int t = e.getKey();
			Double prob = e.getValue();

			if (t == s) {
				continue;
			}

			// reweight probability:
			prob = prob / (1 - selfProb);

			OptimalChoice ch = opt.getFor(r, t);

			if (ch == null) {
				throw new PrismException("Value not computed yet...");
			}

			p += prob * ch.getP();
			theta += prob * ch.getTheta();
			if (extraVerbose) {
				mc.getLog().println(ch + "  // t = " + t + ", R' = " + r);
				mc.getLog().println("p += " + prob + " * " + ch.getP() + " = " + (prob * ch.getP()) + " // t = " + t);
				mc.getLog().println("theta += " + prob + " * " + ch.getTheta() + " = " + (prob * ch.getTheta()) + " // t = " + t);
			}
		}
		if (verbose) {
		mc.getLog().print("p = " + p);
		mc.getLog().println(", theta = " + theta);
		}

		return new OptimalChoice(s, r, alpha, p, theta);
	}

	private int computeSaturationPointDirect(MDPSparse mdp, double CEupper, MDPRewardsSimple rewards, explicit.StateValues P_M, explicit.StateValues E_M) throws PrismException
	{
		double min_value = Double.POSITIVE_INFINITY;

		StopWatch timer = new StopWatch(mc.getLog());
		timer.start("computing saturation point");

		mc.getLog().println("Computing saturation point");

		for (int s = 0; s < mdp.getNumStates(); s++) {
			double y_s = (double) P_M.getValue(s);
			double theta_s = (double) E_M.getValue(s);

			if (verbose) mc.getLog().println("y_" + s + " = " + y_s + ", theta_" + s + " = " + theta_s);

			for (int alpha = 0; alpha < mdp.getNumChoices(s); alpha++) {
				double y_s_a = 0;
				double theta_s_a = 0;

				for (Iterator<Entry<Integer, Double>> it = mdp.getTransitionsIterator(s, alpha); it.hasNext(); ) {
					Entry<Integer, Double> e = it.next();
					int t = e.getKey();
					Double prob = e.getValue();

					y_s_a += prob * (double)P_M.getValue(t);
					theta_s_a += prob * (double)E_M.getValue(t);

				}

				double r = rewards.getStateReward(s);
				r += rewards.getTransitionReward(s, alpha);
				r *= y_s_a;

				theta_s_a += r;

				if (verbose) mc.getLog().println("y_" + s + "_" + alpha + " = " + y_s_a + ", theta_" + s + "_" + alpha + " = " + theta_s_a);

				if (y_s_a >= y_s) {
					if (verbose) mc.getLog().println("Skipping, y_s_a >= y_s");
					continue;
				}

				double diff = (theta_s - theta_s_a) / (y_s - y_s_a);
				if (verbose) mc.getLog().println("Diff = " + diff);
				if (diff < min_value) {
					min_value = diff;
					if (verbose) mc.getLog().println("New min value = " + diff);
				}
			}
		}

		int saturation_point;
		if (Double.isInfinite(min_value)) {
			// if there were no y_s_a < y_s, then the deterministic memoryless scheduler M
			// is optimal, i.e., we don't need any memory...
			saturation_point = 0;
		} else {
			double r_frac = CEupper - min_value;
			if (r_frac < 0)
				r_frac = 0;   // saturation point can never by < 0
			saturation_point = (int) Math.ceil(r_frac);
			if (saturation_point < r_frac) {
				throw new PrismException("Can not store computed saturation point (next integer >= " + r_frac + " as an integer...");
			}
		}
		timer.stop();

		mc.getLog().println("\nComputed saturation point = " + saturation_point); 
		return saturation_point;
	}

	private BigInteger countMDSchedulers(MDPSparse mdp)
	{
		BigInteger result = BigInteger.valueOf(1);
		for (int s = 0; s < mdp.getNumStates(); s++) {
			result = result.multiply(BigInteger.valueOf(mdp.getNumChoices()));
		}
		return result;
	}

	private boolean hasSelfloopForAction(MDPSparse mdp, int s, int action)
	{
		for (Iterator<Integer> it = mdp.getSuccessorsIterator(s, action); it.hasNext(); ) {
			if (it.next() == s) {
				return true;
			}
		}
		return false;
	}

	private String constructScheduler(OptimalChoicesComputed<OptimalChoice> witness)
	{
		StringBuffer buffer = new StringBuffer();
		buffer.append("Scheduler:\n\n");
		witness.print(buffer);
		return buffer.toString();
	}

	private double verifyScheduled(NondetModel mdp, OptimalChoicesComputed<OptimalChoice> opt, double CEmax) throws PrismException
	{
		mc.getLog().println("=========== Verifying result =========================");

		int Rmax = opt.getRequiredMemory();
		RewardCounterProduct<NondetModel> product = RewardCounterProduct.generate(mc, mdp, mdp.getTransRewards(0).copy(), Rmax, mdp.getStart().copy());

		NondetModel mdpR = product.getTransformedModel();
		JDDNode trans = mdpR.getTrans().copy();

		StateValuesMTBDD reach = new StateValuesMTBDD(mdp.getReach().copy(), mdp);

		JDDNode ddActions = mdpR.getTrans01().copy();
		ddActions = JDD.ThereExists(ddActions, mdpR.getAllDDColVars());
		SanityJDD.checkIsDDOverVars(ddActions, mdpR.getAllDDRowVars(), mdpR.getAllDDNondetVars());

		JDDVars stateActionVars = mdpR.getAllDDRowVars().copy();
		stateActionVars.mergeVarsFrom(mdpR.getAllDDNondetVars());

		for (int r : opt.getLevelsDescending()) {
			for (int s : opt.getStatesForLevel(r)) {
				OptimalChoice choice = opt.getFor(r,s);
				int optChoice = choice.getChoice();
				if (optChoice < 0)
					continue;

				JDDNode ddS = reach.convertIndexToDD(s);
				// JDD.PrintMinterms(mc.getLog(), ddS.copy(), "ddS for s = "+s);
				ddS = JDD.And(ddS, product.getStatesWithAccumulatedReward(r));
				// JDD.PrintMinterms(mc.getLog(), ddS.copy(), "ddS for s = " + s +" r = " + r);

				if (ddS.equals(JDD.ZERO)) {
					JDD.Deref(ddS);
					continue;
				}

				JDDNode ddSAct = JDD.And(ddActions.copy(), ddS.copy());
				// JDD.PrintMinterms(mc.getLog(), ddSAct.copy(), "ddSAct for s = " + s +" r = " + r);

				SanityJDD.checkIsDDOverVars(ddSAct, mdpR.getAllDDRowVars(), mdpR.getAllDDNondetVars());
				ODDNode odd = ODDUtils.BuildODD(ddSAct, stateActionVars);
				JDDNode ddAct = ODDUtils.SingleIndexToDD(optChoice, odd, stateActionVars);
				// JDD.PrintMinterms(mc.getLog(), ddAct.copy(), "ddAct for s = " + s +" r = " + r + " act = " +optChoice);
				JDDNode ddMask = JDD.Or(JDD.Not(ddS.copy()), ddAct);
				// JDD.PrintMinterms(mc.getLog(), ddMask.copy(), "ddMask for s = " + s +" r = " + r + " act = " +optChoice);

				// JDD.PrintMinterms(mc.getLog(), trans.copy(), "trans(1) for s = " + s +" r = " + r + " act = " +optChoice);
				trans = JDD.Times(trans, ddMask);
				// JDD.PrintMinterms(mc.getLog(), trans.copy(), "trans(2) for s = " + s +" r = " + r + " act = " +optChoice);
				
				JDD.Deref(ddS, ddSAct);
			}
		}

		reach.clear();
		JDD.Deref(ddActions);
		stateActionVars.derefAll();

		final JDDNode transDone = trans;

		NondetModelTransformation transform = new NondetModelTransformation(mdpR) {
			@Override
			public int getExtraStateVariableCount() {return 0;}

			@Override
			public int getExtraActionVariableCount() {return 0;}

			@Override
			public JDDNode getTransformedTrans() throws PrismException
			{
				return transDone.copy();
			}

			@Override
			public JDDNode getTransformedStart() throws PrismException
			{
				return originalModel.getStart().copy();
			}

			public void clear()
			{
				super.clear();
			}
		};

		NondetModel transformed = mdpR.getTransformed(transform);
		if (verbose) {
			try {
				transformed.exportToFile(Prism.EXPORT_DOT, true, new File("scheduled.dot"));
				transformed.exportStateRewardsToFile(Prism.EXPORT_PLAIN, new File("scheduled.srew"));
				transformed.exportTransRewardsToFile(Prism.EXPORT_PLAIN, true, new File("scheduled.trew"));
			} catch (FileNotFoundException e) {}
		}
		
		ProbModel dtmc = NondetModel.toDTMC(transformed);
		if (verbose) {
			try {
				dtmc.exportToFile(Prism.EXPORT_DOT, true, new File("scheduled-dtmc.dot"));
				dtmc.exportStateRewardsToFile(Prism.EXPORT_PLAIN, new File("scheduled-dtmc.srew"));
				dtmc.exportTransRewardsToFile(Prism.EXPORT_PLAIN, true, new File("scheduled-dtmc.trew"));
			} catch (FileNotFoundException e) {}
		}

		ProbModelChecker dtmcMC = (ProbModelChecker) mc.createModelChecker(dtmc);
		dtmcMC.setLog(new PrismDevNullLog());
		ExpressionTemporal ePath = new ExpressionTemporal(ExpressionTemporal.P_F, null, new ExpressionLabel(goalLabel));
		Expression objective = new ExpressionReward(ePath, "=", null);
		ExpressionConditional e = new ExpressionConditional(objective, ePath.deepCopy());

		StateValues value = dtmcMC.checkExpression(e, dtmc.getStart().copy());
		mc.getLog().println("Result verification = ");
		value.print(mc.getLog());

		double result = value.firstFromBDD(dtmc.getStart());

		value.clear();
		dtmc.clear();
		transform.clear();
		transformed.clear();
		product.clear();

		JDD.Deref(transDone);

		return result;
	}
}
