package prism.conditional.transform;

import jdd.JDD;
import jdd.JDDNode;
import jdd.JDDVars;
import jdd.SanityJDD;
import prism.ModelTransformation;
import prism.NondetModel;
import prism.NondetModelTransformation;
import prism.PrismComponent;
import prism.PrismException;
import prism.StateValues;

public class MDPAddTauToFail extends NondetModelTransformation
{
	public static final boolean ROW    = true;
	public static final boolean COLUMN = false;

	private JDDNode tauStates;
	private JDDNode failState;
	private boolean needNewFailState;
	private PrismComponent parent;
	private boolean verbose;

	private MDPAddTauToFail(PrismComponent parent, NondetModel model, JDDNode tauStates, JDDNode failState)
	{
		super(model);

		this.parent = parent;

		this.tauStates = tauStates;
		if (failState == null) {
			failState = JDD.Constant(0);
		}
		JDDNode fail = JDD.And(failState.copy(), model.getReach().copy());
		if (fail.equals(JDD.ZERO)) {
			needNewFailState = true;
			this.failState = null;
			JDD.Deref(fail);
			if (failState != null)
				JDD.Deref(failState);
		} else {
			needNewFailState = false;
			this.failState = fail;
			JDD.Deref(failState);
			SanityJDD.check(JDD.isSingleton(fail, model.getAllDDRowVars()), "fail state is not a singleton");
		}
		
		SanityJDD.checkIsContainedIn(tauStates, model.getReach());
	}

	public void clear()
	{
		super.clear();

		JDD.Deref(tauStates);
		if (failState != null) JDD.Deref(failState);
	}

	private JDDNode allZero(JDDVars vars)
	{
		JDDNode result = JDD.Constant(1);
		for (JDDNode var : vars) {
			result = JDD.And(result, JDD.Not(var.copy()));
		}
		return result;
	}
	
	private JDDNode getFailState(boolean row)
	{
		if (needNewFailState) {
			JDDNode fail = row ? extraRowVars.getVar(0).copy() : extraColVars.getVar(0).copy();
			fail = JDD.And(fail, allZero(row ? originalModel.getAllDDRowVars() : originalModel.getAllDDColVars()));
			return fail;
		} else {
			return row ? failState.copy() : JDD.PermuteVariables(failState.copy(), originalModel.getAllDDRowVars(), originalModel.getAllDDColVars());
		}
	}

	private JDDNode normal(boolean row)
	{
		if (needNewFailState) {
			 return JDD.Not(row ? extraRowVars.getVar(0).copy() : extraColVars.getVar(0).copy());
		} else {
			return JDD.Constant(1);
		}
	}

	private JDDNode tau()
	{
		return JDD.And(extraActionVars.getVar(0).copy(), allZero(originalModel.getAllDDNondetVars()));
	}

	private JDDNode notTau()
	{
		return JDD.Not(extraActionVars.getVar(0).copy());
	}

	@Override
	public int getExtraStateVariableCount()
	{
		return needNewFailState ? 1 : 0;
	}

	@Override
	public int getExtraActionVariableCount()
	{
		return 1;
	}

	@Override
	public JDDNode getTransformedTrans() throws PrismException
	{
		if (verbose) JDD.PrintMinterms(parent.getLog(), getFailState(ROW), "failState");
		if (verbose) JDD.PrintMinterms(parent.getLog(), tauStates.copy(), "tauStates");
		
		JDDNode newTrans = originalModel.getTrans().copy();
		if (verbose) JDD.PrintMinterms(parent.getLog(), newTrans.copy(), "trans");
		if (!needNewFailState) {
			// strip all transitions from fail state
			newTrans = JDD.Times(newTrans, JDD.Not(failState.copy()));
			if (verbose) JDD.PrintMinterms(parent.getLog(), newTrans.copy(), "newTrans (after strip)");
		} else {
			newTrans = JDD.Times(newTrans, normal(ROW), normal(COLUMN));
			if (verbose) JDD.PrintMinterms(parent.getLog(), newTrans.copy(), "newTrans (after normal)");
		}
		newTrans = JDD.Times(newTrans, notTau());
		if (verbose) JDD.PrintMinterms(parent.getLog(), newTrans.copy(), "newTrans (after notTau)");

		JDDNode tauFromTauStates = JDD.Times(tauStates.copy(), normal(ROW), tau(), getFailState(COLUMN));
		if (verbose) JDD.PrintMinterms(parent.getLog(), tauFromTauStates.copy(), "tauFromTauStates");
		JDDNode failSelfLoop = JDD.Times(getFailState(ROW), getFailState(COLUMN), tau());
		if (verbose) JDD.PrintMinterms(parent.getLog(), failSelfLoop.copy(), "failSelfLoop");

		newTrans = JDD.Apply(JDD.MAX, newTrans, tauFromTauStates);
		if (verbose) JDD.PrintMinterms(parent.getLog(), newTrans.copy(), "newTrans (after normal)");
		newTrans = JDD.Apply(JDD.MAX, newTrans, failSelfLoop);
		if (verbose) JDD.PrintMinterms(parent.getLog(), newTrans.copy(), "newTrans (after normal)");

		return newTrans;
	}

	@Override
	public JDDNode getTransformedStart() throws PrismException
	{
		return JDD.And(originalModel.getStart().copy(), normal(ROW));
	}

	@Override
	public JDDNode getReachableStates()
	{
		return JDD.Or(JDD.And(originalModel.getReach().copy(), normal(ROW)),
		              getFailState(ROW));
	}

	@Override
	public JDDNode getTransformedStateReward(JDDNode rew)
	{
		return JDD.Times(rew.copy(), normal(ROW));
	}
	
	@Override
	public JDDNode getTransformedTransReward(JDDNode rew)
	{
		return JDD.Times(rew.copy(), normal(ROW), normal(COLUMN));
	}

	@Override
	public JDDNode getTransformedTransActions()
	{
		if (originalModel.getTransActions() == null) return null;
		return JDD.Times(originalModel.getTransActions().copy(), normal(ROW));
	}

	public static class MDPWithFail implements ModelTransformation<NondetModel, NondetModel>
	{
		private NondetModel originalModel;
		private NondetModel transformedModel;
		private MDPAddTauToFail transform;

		private MDPWithFail(NondetModel originalModel, NondetModel transformedModel, MDPAddTauToFail transform)
		{
			this.originalModel = originalModel;
			this.transformedModel = transformedModel;
			this.transform = transform;
		}
		
		@Override
		public NondetModel getOriginalModel()
		{
			return originalModel;
		}

		@Override
		public NondetModel getTransformedModel()
		{
			return transformedModel;
		}

		@Override
		public void clear()
		{
			transformedModel.clear();
			transform.clear();
		}

		@Override
		public StateValues projectToOriginalModel(StateValues svTransformedModel) throws PrismException
		{
			svTransformedModel.filter(transform.normal(ROW));
			StateValues result = svTransformedModel.sumOverDDVars(transform.extraRowVars, transform.originalModel);

			svTransformedModel.clear();
			return result;
		}

		@Override
		public JDDNode getTransformedStatesOfInterest()
		{
			throw new RuntimeException("not implemented");
		}

		public JDDNode getFailState()
		{
			return transform.getFailState(ROW);
		}
	};

	public static MDPWithFail transform(PrismComponent parent, NondetModel model, JDDNode tauStates, JDDNode failState) throws PrismException
	{
		MDPAddTauToFail transform = new MDPAddTauToFail(parent, model, tauStates, failState);
		NondetModel transformed = model.getTransformed(transform);

		return new MDPWithFail(model, transformed, transform);
	}
}
